from __future__ import annotations

import atexit
import os
import socket
import threading
import time
from typing import Optional

from lmnr.sdk.log import get_default_logger

from lmnr_claude_code_proxy import run_server, set_current_trace, stop_server

logger = get_default_logger(__name__)

DEFAULT_ANTHROPIC_BASE_URL = "https://api.anthropic.com"
DEFAULT_CC_PROXY_PORT = 45667
CC_PROXY_PORT_ATTEMPTS = 5

_CC_PROXY_LOCK = threading.Lock()
_CC_PROXY_PORT: int | None = None
_CC_PROXY_BASE_URL: str | None = None
_CC_PROXY_TARGET_URL: str | None = None
_CC_PROXY_SHUTDOWN_REGISTERED = False


def _find_available_port(start_port: int, attempts: int) -> Optional[int]:
    for offset in range(attempts):
        candidate = start_port + offset
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                sock.bind(("127.0.0.1", candidate))
            except OSError:
                continue
        return candidate
    return None


def _wait_for_port(port: int, timeout: float = 5.0) -> bool:
    deadline = time.monotonic() + timeout
    while time.monotonic() < deadline:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(0.2)
            try:
                sock.connect(("127.0.0.1", port))
                return True
            except OSError:
                time.sleep(0.1)
    return False


def _stop_cc_proxy_locked():
    global _CC_PROXY_PORT, _CC_PROXY_BASE_URL

    try:
        stop_server()
    except Exception as e:
        logger.debug("Unable to stop cc-proxy: %s", e)

    if _CC_PROXY_TARGET_URL:
        os.environ["ANTHROPIC_BASE_URL"] = _CC_PROXY_TARGET_URL

    _CC_PROXY_PORT = None
    _CC_PROXY_BASE_URL = None


def _stop_cc_proxy():
    with _CC_PROXY_LOCK:
        _stop_cc_proxy_locked()


def _register_proxy_shutdown():
    global _CC_PROXY_SHUTDOWN_REGISTERED
    if not _CC_PROXY_SHUTDOWN_REGISTERED:
        atexit.register(_stop_cc_proxy)
        _CC_PROXY_SHUTDOWN_REGISTERED = True


def get_cc_proxy_base_url() -> str | None:
    return _CC_PROXY_BASE_URL


def start_proxy() -> Optional[str]:
    with _CC_PROXY_LOCK:
        global _CC_PROXY_PORT, _CC_PROXY_BASE_URL, _CC_PROXY_TARGET_URL

        port = _find_available_port(DEFAULT_CC_PROXY_PORT, CC_PROXY_PORT_ATTEMPTS)
        if port is None:
            logger.warning("Unable to allocate port for cc-proxy.")
            return None

        target_url = (
            _CC_PROXY_TARGET_URL
            or os.environ.get("ANTHROPIC_ORIGINAL_BASE_URL")
            or os.environ.get("ANTHROPIC_BASE_URL")
            or DEFAULT_ANTHROPIC_BASE_URL
        )
        _CC_PROXY_TARGET_URL = target_url
        os.environ.setdefault("ANTHROPIC_ORIGINAL_BASE_URL", target_url)

        try:
            run_server(target_url, port=port)
        except OSError as exc:  # pragma: no cover
            logger.warning("Unable to start cc-proxy: %s", exc)
            return None

        if not _wait_for_port(port):
            logger.warning("cc-proxy failed to start on port %s", port)
            stop_server()
            return None

        proxy_base_url = f"http://127.0.0.1:{port}"
        _CC_PROXY_PORT = port
        _CC_PROXY_BASE_URL = proxy_base_url
        os.environ["ANTHROPIC_BASE_URL"] = proxy_base_url
        _register_proxy_shutdown()

        logger.info("Started claude proxy server on: " + str(proxy_base_url))
        return proxy_base_url


def release_proxy() -> None:
    with _CC_PROXY_LOCK:
        _stop_cc_proxy_locked()
        logger.debug("Released claude proxy server")


def set_trace_to_proxy(
    trace_id: str,
    span_id: str,
    project_api_key: str,
    span_path: list[str] = [],
    span_ids_path: list[str] = [],
    laminar_url: str = "https://api.lmnr.ai",
):
    set_current_trace(
        trace_id=trace_id,
        span_id=span_id,
        project_api_key=project_api_key,
        span_path=span_path,
        span_ids_path=span_ids_path,
        laminar_url=laminar_url,
    )



from lmnr.sdk.client.asynchronous.resources.browser_events import AsyncBrowserEvents
from lmnr.sdk.client.asynchronous.resources.evals import AsyncEvals
from lmnr.sdk.client.asynchronous.resources.tags import AsyncTags
from lmnr.sdk.client.asynchronous.resources.evaluators import AsyncEvaluators

__all__ = ["AsyncEvals", "AsyncBrowserEvents", "AsyncTags", "AsyncEvaluators"]



"""OpenTelemetry Groq instrumentation"""

import logging
import os
import time
from typing import Callable, Collection, Union

from opentelemetry import context as context_api
from opentelemetry._events import EventLogger, get_event_logger
from .config import Config
from .event_emitter import (
    emit_choice_events,
    emit_message_events,
    emit_streaming_response_events,
)
from .span_utils import (
    set_input_attributes,
    set_model_input_attributes,
    set_model_response_attributes,
    set_model_streaming_response_attributes,
    set_response_attributes,
    set_streaming_response_attributes,
)
from .utils import (
    error_metrics_attributes,
    shared_metrics_attributes,
    should_emit_events,
)
from .version import __version__
from lmnr.opentelemetry_lib.tracing.context import get_current_context
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY, unwrap
from opentelemetry.metrics import Counter, Histogram, Meter, get_meter
from opentelemetry.semconv_ai import (
    SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY,
    LLMRequestTypeValues,
    Meters,
    SpanAttributes,
)
from opentelemetry.trace import SpanKind, Tracer, get_tracer
from opentelemetry.trace.status import Status, StatusCode
from wrapt import wrap_function_wrapper

from groq._streaming import AsyncStream, Stream

logger = logging.getLogger(__name__)

_instruments = ("groq >= 0.9.0",)


WRAPPED_METHODS = [
    {
        "package": "groq.resources.chat.completions",
        "object": "Completions",
        "method": "create",
        "span_name": "groq.chat",
    },
]
WRAPPED_AMETHODS = [
    {
        "package": "groq.resources.chat.completions",
        "object": "AsyncCompletions",
        "method": "create",
        "span_name": "groq.chat",
    },
]


def is_streaming_response(response):
    return isinstance(response, Stream) or isinstance(response, AsyncStream)


def _with_chat_telemetry_wrapper(func):
    """Helper for providing tracer for wrapper functions. Includes metric collectors."""

    def _with_chat_telemetry(
        tracer,
        token_histogram,
        choice_counter,
        duration_histogram,
        event_logger,
        to_wrap,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                tracer,
                token_histogram,
                choice_counter,
                duration_histogram,
                event_logger,
                to_wrap,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_chat_telemetry


def _create_metrics(meter: Meter):
    token_histogram = meter.create_histogram(
        name=Meters.LLM_TOKEN_USAGE,
        unit="token",
        description="Measures number of input and output tokens used",
    )

    choice_counter = meter.create_counter(
        name=Meters.LLM_GENERATION_CHOICES,
        unit="choice",
        description="Number of choices returned by chat completions call",
    )

    duration_histogram = meter.create_histogram(
        name=Meters.LLM_OPERATION_DURATION,
        unit="s",
        description="GenAI operation duration",
    )

    return token_histogram, choice_counter, duration_histogram


def _process_streaming_chunk(chunk):
    """Extract content, finish_reason and usage from a streaming chunk."""
    if not chunk.choices:
        return None, None, None

    delta = chunk.choices[0].delta
    content = delta.content if hasattr(delta, "content") else None
    finish_reason = chunk.choices[0].finish_reason

    # Extract usage from x_groq if present in the final chunk
    usage = None
    if hasattr(chunk, "x_groq") and chunk.x_groq and chunk.x_groq.usage:
        usage = chunk.x_groq.usage

    return content, finish_reason, usage


def _handle_streaming_response(
    span, accumulated_content, finish_reason, usage, event_logger
):
    set_model_streaming_response_attributes(span, usage)
    if should_emit_events() and event_logger:
        emit_streaming_response_events(accumulated_content, finish_reason, event_logger)
    else:
        set_streaming_response_attributes(
            span, accumulated_content, finish_reason, usage
        )


def _create_stream_processor(response, span, event_logger):
    """Create a generator that processes a stream while collecting telemetry."""
    accumulated_content = ""
    finish_reason = None
    usage = None

    for chunk in response:
        content, chunk_finish_reason, chunk_usage = _process_streaming_chunk(chunk)
        if content:
            accumulated_content += content
        if chunk_finish_reason:
            finish_reason = chunk_finish_reason
        if chunk_usage:
            usage = chunk_usage
        yield chunk

    _handle_streaming_response(
        span, accumulated_content, finish_reason, usage, event_logger
    )

    if span.is_recording():
        span.set_status(Status(StatusCode.OK))

    span.end()


async def _create_async_stream_processor(response, span, event_logger):
    """Create an async generator that processes a stream while collecting telemetry."""
    accumulated_content = ""
    finish_reason = None
    usage = None

    async for chunk in response:
        content, chunk_finish_reason, chunk_usage = _process_streaming_chunk(chunk)
        if content:
            accumulated_content += content
        if chunk_finish_reason:
            finish_reason = chunk_finish_reason
        if chunk_usage:
            usage = chunk_usage
        yield chunk

    _handle_streaming_response(
        span, accumulated_content, finish_reason, usage, event_logger
    )

    if span.is_recording():
        span.set_status(Status(StatusCode.OK))

    span.end()


def _handle_input(span, kwargs, event_logger):
    set_model_input_attributes(span, kwargs)
    if should_emit_events() and event_logger:
        emit_message_events(kwargs, event_logger)
    else:
        set_input_attributes(span, kwargs)


def _handle_response(span, response, token_histogram, event_logger):
    set_model_response_attributes(span, response, token_histogram)
    if should_emit_events() and event_logger:
        emit_choice_events(response, event_logger)
    else:
        set_response_attributes(span, response)


@_with_chat_telemetry_wrapper
def _wrap(
    tracer: Tracer,
    token_histogram: Histogram,
    choice_counter: Counter,
    duration_histogram: Histogram,
    event_logger: Union[EventLogger, None],
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """Instruments and calls every function defined in TO_WRAP."""
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return wrapped(*args, **kwargs)

    name = to_wrap.get("span_name")
    span = tracer.start_span(
        name,
        kind=SpanKind.CLIENT,
        attributes={
            SpanAttributes.LLM_SYSTEM: "Groq",
            SpanAttributes.LLM_REQUEST_TYPE: LLMRequestTypeValues.COMPLETION.value,
        },
        context=get_current_context(),
    )

    _handle_input(span, kwargs, event_logger)

    start_time = time.time()
    try:
        response = wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        end_time = time.time()
        attributes = error_metrics_attributes(e)

        if duration_histogram:
            duration = end_time - start_time
            duration_histogram.record(duration, attributes=attributes)

        raise e

    end_time = time.time()

    if is_streaming_response(response):
        try:
            return _create_stream_processor(response, span, event_logger)
        except Exception as ex:
            logger.warning(
                "Failed to process streaming response for groq span, error: %s",
                str(ex),
            )
            span.set_status(Status(StatusCode.ERROR))
            span.end()
            raise
    elif response:
        try:
            metric_attributes = shared_metrics_attributes(response)

            if duration_histogram:
                duration = time.time() - start_time
                duration_histogram.record(
                    duration,
                    attributes=metric_attributes,
                )

            _handle_response(span, response, token_histogram, event_logger)

        except Exception as ex:  # pylint: disable=broad-except
            logger.warning(
                "Failed to set response attributes for groq span, error: %s",
                str(ex),
            )

        if span.is_recording():
            span.set_status(Status(StatusCode.OK))
    span.end()
    return response


@_with_chat_telemetry_wrapper
async def _awrap(
    tracer,
    token_histogram: Histogram,
    choice_counter: Counter,
    duration_histogram: Histogram,
    event_logger: Union[EventLogger, None],
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """Instruments and calls every function defined in TO_WRAP."""
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return await wrapped(*args, **kwargs)

    name = to_wrap.get("span_name")
    span = tracer.start_span(
        name,
        kind=SpanKind.CLIENT,
        attributes={
            SpanAttributes.LLM_SYSTEM: "Groq",
            SpanAttributes.LLM_REQUEST_TYPE: LLMRequestTypeValues.COMPLETION.value,
        },
        context=get_current_context(),
    )

    _handle_input(span, kwargs, event_logger)

    start_time = time.time()

    try:
        response = await wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        end_time = time.time()
        attributes = error_metrics_attributes(e)

        if duration_histogram:
            duration = end_time - start_time
            duration_histogram.record(duration, attributes=attributes)

        raise e

    end_time = time.time()

    if is_streaming_response(response):
        try:
            return await _create_async_stream_processor(response, span, event_logger)
        except Exception as ex:
            logger.warning(
                "Failed to process streaming response for groq span, error: %s",
                str(ex),
            )
            span.set_status(Status(StatusCode.ERROR))
            span.end()
            raise
    elif response:
        metric_attributes = shared_metrics_attributes(response)

        if duration_histogram:
            duration = time.time() - start_time
            duration_histogram.record(
                duration,
                attributes=metric_attributes,
            )

        _handle_response(span, response, token_histogram, event_logger)

        if span.is_recording():
            span.set_status(Status(StatusCode.OK))
    span.end()
    return response


def is_metrics_enabled() -> bool:
    return (os.getenv("TRACELOOP_METRICS_ENABLED") or "true").lower() == "true"


class GroqInstrumentor(BaseInstrumentor):
    """An instrumentor for Groq's client library."""

    def __init__(
        self,
        enrich_token_usage: bool = False,
        exception_logger=None,
        use_legacy_attributes: bool = True,
        get_common_metrics_attributes: Callable[[], dict] = lambda: {},
    ):
        super().__init__()
        Config.exception_logger = exception_logger
        Config.enrich_token_usage = enrich_token_usage
        Config.get_common_metrics_attributes = get_common_metrics_attributes
        Config.use_legacy_attributes = use_legacy_attributes

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        # meter and counters are inited here
        meter_provider = kwargs.get("meter_provider")
        meter = get_meter(__name__, __version__, meter_provider)

        if is_metrics_enabled():
            (
                token_histogram,
                choice_counter,
                duration_histogram,
            ) = _create_metrics(meter)
        else:
            (
                token_histogram,
                choice_counter,
                duration_histogram,
            ) = (None, None, None)

        event_logger = None
        if not Config.use_legacy_attributes:
            event_logger_provider = kwargs.get("event_logger_provider")
            event_logger = get_event_logger(
                __name__, __version__, event_logger_provider=event_logger_provider
            )

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")

            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _wrap(
                        tracer,
                        token_histogram,
                        choice_counter,
                        duration_histogram,
                        event_logger,
                        wrapped_method,
                    ),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _awrap(
                        tracer,
                        token_histogram,
                        choice_counter,
                        duration_histogram,
                        event_logger,
                        wrapped_method,
                    ),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrapped_method.get("method"),
            )
        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrapped_method.get("method"),
            )



import logging
import time
from typing import Optional

from opentelemetry._events import EventLogger
from .config import Config
from .event_emitter import (
    emit_streaming_response_events,
)
from .span_utils import (
    set_streaming_response_attributes,
)
from .utils import (
    count_prompt_tokens_from_request,
    dont_throw,
    error_metrics_attributes,
    set_span_attribute,
    shared_metrics_attributes,
    should_emit_events,
)
from opentelemetry.metrics import Counter, Histogram
from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import (
    GEN_AI_RESPONSE_ID,
)
from opentelemetry.semconv_ai import SpanAttributes
from opentelemetry.trace.status import Status, StatusCode

logger = logging.getLogger(__name__)


@dont_throw
def _process_response_item(item, complete_response):
    if item.type == "message_start":
        complete_response["model"] = item.message.model
        usage = dict(item.message.usage)
        complete_response["usage"] = usage
        complete_response["service_tier"] = usage.get("service_tier") or None
        complete_response["id"] = item.message.id
    elif item.type == "content_block_start":
        index = item.index
        if len(complete_response.get("events")) <= index:
            complete_response["events"].append(
                {"index": index, "text": "", "type": item.content_block.type}
            )
            if item.content_block.type == "tool_use":
                complete_response["events"][index]["id"] = item.content_block.id
                complete_response["events"][index]["name"] = item.content_block.name
                complete_response["events"][index]["input"] = ""

    elif item.type == "content_block_delta":
        index = item.index
        if item.delta.type == "thinking_delta":
            complete_response["events"][index]["text"] += item.delta.thinking or ""
        elif item.delta.type == "text_delta":
            complete_response["events"][index]["text"] += item.delta.text or ""
        elif item.delta.type == "input_json_delta":
            complete_response["events"][index]["input"] += item.delta.partial_json
    elif item.type == "message_delta":
        for event in complete_response.get("events", []):
            event["finish_reason"] = item.delta.stop_reason
        if item.usage:
            if "usage" in complete_response:
                item_output_tokens = dict(item.usage).get("output_tokens", 0)
                existing_output_tokens = complete_response["usage"].get(
                    "output_tokens", 0
                )
                complete_response["usage"]["output_tokens"] = (
                    item_output_tokens + existing_output_tokens
                )
            else:
                complete_response["usage"] = dict(item.usage)
    elif item.type in ["message_stop", "message_start"]:
        # raw stream returns the service_tier in the message_start event
        # messages.stream returns the service_tier in the message_stop event
        usage = dict(item.message.usage or {})
        complete_response["service_tier"] = usage.get("service_tier")


def _set_token_usage(
    span,
    complete_response,
    prompt_tokens,
    completion_tokens,
    metric_attributes: dict = {},
    token_histogram: Histogram = None,
    choice_counter: Counter = None,
):
    cache_read_tokens = (
        complete_response.get("usage", {}).get("cache_read_input_tokens", 0) or 0
    )
    cache_creation_tokens = (
        complete_response.get("usage", {}).get("cache_creation_input_tokens", 0) or 0
    )

    input_tokens = prompt_tokens + cache_read_tokens + cache_creation_tokens
    total_tokens = input_tokens + completion_tokens

    set_span_attribute(span, SpanAttributes.LLM_USAGE_PROMPT_TOKENS, input_tokens)
    set_span_attribute(
        span, SpanAttributes.LLM_USAGE_COMPLETION_TOKENS, completion_tokens
    )
    set_span_attribute(span, SpanAttributes.LLM_USAGE_TOTAL_TOKENS, total_tokens)

    set_span_attribute(
        span, SpanAttributes.LLM_RESPONSE_MODEL, complete_response.get("model")
    )
    set_span_attribute(
        span, SpanAttributes.LLM_USAGE_CACHE_READ_INPUT_TOKENS, cache_read_tokens
    )
    set_span_attribute(
        span,
        SpanAttributes.LLM_USAGE_CACHE_CREATION_INPUT_TOKENS,
        cache_creation_tokens,
    )

    if token_histogram and type(input_tokens) is int and input_tokens >= 0:
        token_histogram.record(
            input_tokens,
            attributes={
                **metric_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: "input",
            },
        )

    if token_histogram and type(completion_tokens) is int and completion_tokens >= 0:
        token_histogram.record(
            completion_tokens,
            attributes={
                **metric_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: "output",
            },
        )

    if type(complete_response.get("events")) is list and choice_counter:
        for event in complete_response.get("events"):
            choice_counter.add(
                1,
                attributes={
                    **metric_attributes,
                    SpanAttributes.LLM_RESPONSE_FINISH_REASON: event.get(
                        "finish_reason"
                    ),
                },
            )


def _handle_streaming_response(span, event_logger, complete_response):
    if should_emit_events() and event_logger:
        emit_streaming_response_events(event_logger, complete_response)
    else:
        if not span.is_recording():
            return
        set_streaming_response_attributes(span, complete_response.get("events"))


@dont_throw
def build_from_streaming_response(
    span,
    response,
    instance,
    start_time,
    token_histogram: Histogram = None,
    choice_counter: Counter = None,
    duration_histogram: Histogram = None,
    exception_counter: Counter = None,
    event_logger: Optional[EventLogger] = None,
    kwargs: dict = {},
):
    complete_response = {
        "events": [],
        "model": "",
        "usage": {},
        "id": "",
        "service_tier": None,
    }

    for item in response:
        try:
            yield item
        except Exception as e:
            attributes = error_metrics_attributes(e)
            if exception_counter:
                exception_counter.add(1, attributes=attributes)
            raise e
        _process_response_item(item, complete_response)

    metric_attributes = shared_metrics_attributes(complete_response)
    set_span_attribute(span, GEN_AI_RESPONSE_ID, complete_response.get("id"))
    set_span_attribute(
        span,
        "anthropic.response.service_tier",
        complete_response.get("service_tier"),
    )
    if duration_histogram:
        duration = time.time() - start_time
        duration_histogram.record(
            duration,
            attributes=metric_attributes,
        )

    # calculate token usage
    if Config.enrich_token_usage:
        try:
            completion_tokens = -1
            # prompt_usage
            if usage := complete_response.get("usage"):
                prompt_tokens = usage.get("input_tokens", 0) or 0
            else:
                prompt_tokens = count_prompt_tokens_from_request(instance, kwargs)

            # completion_usage
            if usage := complete_response.get("usage"):
                completion_tokens = usage.get("output_tokens", 0) or 0
            else:
                completion_content = ""
                if complete_response.get("events"):
                    model_name = complete_response.get("model") or None
                    for event in complete_response.get("events"):
                        if event.get("text"):
                            completion_content += event.get("text")

                    if model_name and hasattr(instance, "count_tokens"):
                        completion_tokens = instance.count_tokens(completion_content)

            _set_token_usage(
                span,
                complete_response,
                prompt_tokens,
                completion_tokens,
                metric_attributes,
                token_histogram,
                choice_counter,
            )
        except Exception as e:
            logger.warning("Failed to set token usage, error: %s", e)

    _handle_streaming_response(span, event_logger, complete_response)

    if span.is_recording():
        span.set_status(Status(StatusCode.OK))
        span.end()


@dont_throw
async def abuild_from_streaming_response(
    span,
    response,
    instance,
    start_time,
    token_histogram: Histogram = None,
    choice_counter: Counter = None,
    duration_histogram: Histogram = None,
    exception_counter: Counter = None,
    event_logger: Optional[EventLogger] = None,
    kwargs: dict = {},
):
    complete_response = {
        "events": [],
        "model": "",
        "usage": {},
        "id": "",
        "service_tier": None,
    }
    async for item in response:
        try:
            yield item
        except Exception as e:
            attributes = error_metrics_attributes(e)
            if exception_counter:
                exception_counter.add(1, attributes=attributes)
            raise e
        _process_response_item(item, complete_response)

    set_span_attribute(span, GEN_AI_RESPONSE_ID, complete_response.get("id"))
    set_span_attribute(
        span,
        "anthropic.response.service_tier",
        complete_response.get("service_tier"),
    )

    metric_attributes = shared_metrics_attributes(complete_response)

    if duration_histogram:
        duration = time.time() - start_time
        duration_histogram.record(
            duration,
            attributes=metric_attributes,
        )

    # calculate token usage
    if Config.enrich_token_usage:
        try:
            # prompt_usage
            if usage := complete_response.get("usage"):
                prompt_tokens = usage.get("input_tokens", 0)
            else:
                prompt_tokens = count_prompt_tokens_from_request(instance, kwargs)

            # completion_usage
            if usage := complete_response.get("usage"):
                completion_tokens = usage.get("output_tokens", 0)
            else:
                completion_content = ""
                if complete_response.get("events"):
                    model_name = complete_response.get("model") or None
                    for event in complete_response.get("events"):
                        if event.get("text"):
                            completion_content += event.get("text")

                    if model_name and hasattr(instance, "count_tokens"):
                        completion_tokens = instance.count_tokens(completion_content)

            _set_token_usage(
                span,
                complete_response,
                prompt_tokens,
                completion_tokens,
                metric_attributes,
                token_histogram,
                choice_counter,
            )
        except Exception as e:
            logger.warning("Failed to set token usage, error: %s", str(e))

    _handle_streaming_response(span, event_logger, complete_response)

    if span.is_recording():
        span.set_status(Status(StatusCode.OK))
        span.end()


class WrappedMessageStreamManager:
    """Wrapper for MessageStreamManager that handles instrumentation"""

    def __init__(
        self,
        stream_manager,
        span,
        instance,
        start_time,
        token_histogram,
        choice_counter,
        duration_histogram,
        exception_counter,
        event_logger,
        kwargs,
    ):
        self._stream_manager = stream_manager
        self._span = span
        self._instance = instance
        self._start_time = start_time
        self._token_histogram = token_histogram
        self._choice_counter = choice_counter
        self._duration_histogram = duration_histogram
        self._exception_counter = exception_counter
        self._event_logger = event_logger
        self._kwargs = kwargs

    def __enter__(self):
        # Call the original stream manager's __enter__ to get the actual stream
        stream = self._stream_manager.__enter__()
        # Return the wrapped stream
        return build_from_streaming_response(
            self._span,
            stream,
            self._instance,
            self._start_time,
            self._token_histogram,
            self._choice_counter,
            self._duration_histogram,
            self._exception_counter,
            self._event_logger,
            self._kwargs,
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._stream_manager.__exit__(exc_type, exc_val, exc_tb)


class WrappedAsyncMessageStreamManager:
    """Wrapper for AsyncMessageStreamManager that handles instrumentation"""

    def __init__(
        self,
        stream_manager,
        span,
        instance,
        start_time,
        token_histogram,
        choice_counter,
        duration_histogram,
        exception_counter,
        event_logger,
        kwargs,
    ):
        self._stream_manager = stream_manager
        self._span = span
        self._instance = instance
        self._start_time = start_time
        self._token_histogram = token_histogram
        self._choice_counter = choice_counter
        self._duration_histogram = duration_histogram
        self._exception_counter = exception_counter
        self._event_logger = event_logger
        self._kwargs = kwargs

    async def __aenter__(self):
        # Call the original stream manager's __aenter__ to get the actual stream
        stream = await self._stream_manager.__aenter__()
        # Return the wrapped stream
        return abuild_from_streaming_response(
            self._span,
            stream,
            self._instance,
            self._start_time,
            self._token_histogram,
            self._choice_counter,
            self._duration_histogram,
            self._exception_counter,
            self._event_logger,
            self._kwargs,
        )

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self._stream_manager.__aexit__(exc_type, exc_val, exc_tb)



"""Base class for resource objects."""

import httpx


class BaseResource:
    """Base class for all API resources."""

    def __init__(self, client: httpx.Client, base_url: str, project_api_key: str):
        """Initialize the resource.

        Args:
            client (httpx.Client): HTTP client instance
            base_url (str): Base URL for the API
            project_api_key (str): Project API key
        """
        self._client = client
        self._base_url = base_url
        self._project_api_key = project_api_key

    def _headers(self) -> dict[str, str]:
        """Generate request headers with authentication.

        Returns:
            dict[str, str]: Headers dictionary
        """
        assert self._project_api_key is not None, "Project API key is not set"
        return {
            "Authorization": "Bearer " + self._project_api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
        }



"""OpenTelemetry CUA instrumentation"""

import logging
from typing import Collection

from lmnr.sdk.utils import get_input_from_func_args, json_dumps
from lmnr import Laminar
from lmnr.opentelemetry_lib.tracing.context import get_current_context
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap

from opentelemetry import trace
from opentelemetry.trace import Span
from opentelemetry.trace.status import Status, StatusCode
from wrapt import wrap_function_wrapper

from .utils import payload_to_placeholder

logger = logging.getLogger(__name__)

_instruments = ("cua-computer >= 0.4.0",)


WRAPPED_METHODS = [
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "close",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "force_close",
    },
]
WRAPPED_AMETHODS = [
    {
        "package": "computer.computer",
        "object": "Computer",
        "method": "__aenter__",
        "action": "start_parent_span",
    },
    {
        "package": "computer.computer",
        "object": "Computer",
        "method": "__aexit__",
        "action": "end_parent_span",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "mouse_down",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "mouse_up",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "left_click",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "right_click",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "double_click",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "move_cursor",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "drag_to",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "drag",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "key_down",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "key_up",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "type_text",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "press",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "hotkey",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "scroll",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "scroll_down",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "scroll_up",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "screenshot",
        "output_formatter": payload_to_placeholder,
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "get_screen_size",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "get_cursor_position",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "copy_to_clipboard",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "set_clipboard",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "file_exists",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "directory_exists",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "list_dir",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "read_text",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "write_text",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "read_bytes",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "write_bytes",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "delete_file",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "create_dir",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "delete_dir",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "get_file_size",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "run_command",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "get_accessibility_tree",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "to_screen_coordinates",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "get_active_window_bounds",
    },
    {
        "package": "computer.interface.generic",
        "object": "GenericComputerInterface",
        "method": "to_screenshot_coordinates",
    },
]


def _with_wrapper(func):
    """Helper for providing tracer for wrapper functions. Includes metric collectors."""

    def wrapper(
        to_wrap,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                to_wrap,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return wrapper


def add_input_to_parent_span(span, instance):
    # api_key is skipped on purpose
    params = {}
    if hasattr(instance, "display"):
        params["display"] = instance.display
    if hasattr(instance, "memory"):
        params["memory"] = instance.memory
    if hasattr(instance, "cpu"):
        params["cpu"] = instance.cpu
    if hasattr(instance, "os_type"):
        params["os_type"] = instance.os_type
    if hasattr(instance, "name"):
        params["name"] = instance.name
    if hasattr(instance, "image"):
        params["image"] = instance.image
    if hasattr(instance, "shared_directories"):
        params["shared_directories"] = instance.shared_directories
    if hasattr(instance, "use_host_computer_server"):
        params["use_host_computer_server"] = instance.use_host_computer_server
    if hasattr(instance, "verbosity"):
        if (
            isinstance(instance.verbosity, int)
            and instance.verbosity in logging._levelToName
        ):
            params["verbosity"] = logging._levelToName[instance.verbosity]
        else:
            params["verbosity"] = instance.verbosity
    if hasattr(instance, "telemetry_enabled"):
        params["telemetry_enabled"] = instance.telemetry_enabled
    if hasattr(instance, "provider_type"):
        params["provider_type"] = instance.provider_type
    if hasattr(instance, "port"):
        params["port"] = instance.port
    if hasattr(instance, "noVNC_port"):
        params["noVNC_port"] = instance.noVNC_port
    if hasattr(instance, "host"):
        params["host"] = instance.host
    if hasattr(instance, "storage"):
        params["storage"] = instance.storage
    if hasattr(instance, "ephemeral"):
        params["ephemeral"] = instance.ephemeral
    if hasattr(instance, "experiments"):
        params["experiments"] = instance.experiments
    span.set_attribute("lmnr.span.input", json_dumps(params))


@_with_wrapper
def _wrap(
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    if to_wrap.get("action") == "start_parent_span":
        parent_span = Laminar.start_span("computer.run")
        add_input_to_parent_span(parent_span, instance)
        result = wrapped(*args, **kwargs)
        try:
            instance._interface._lmnr_parent_span = parent_span
        except Exception:
            pass
        return result
    elif to_wrap.get("action") == "end_parent_span":
        result = wrapped(*args, **kwargs)
        try:
            parent_span: Span = instance._interface._lmnr_parent_span
            if parent_span and parent_span.is_recording():
                parent_span.end()
        except Exception:
            pass
        return result

    # if there's no parent span, use
    parent_span = trace.get_current_span(context=get_current_context())
    try:
        if instance._lmnr_parent_span:
            parent_span: Span = instance._lmnr_parent_span
    except Exception:
        pass

    with Laminar.use_span(parent_span):
        instance_name = "interface"
        with Laminar.start_as_current_span(
            f"{instance_name}.{to_wrap.get('method')}", span_type="TOOL"
        ) as span:
            span.set_attribute(
                "lmnr.span.input",
                json_dumps(get_input_from_func_args(wrapped, True, args, kwargs)),
            )
            try:
                result = wrapped(*args, **kwargs)
            except Exception as e:  # pylint: disable=broad-except
                span.set_status(Status(StatusCode.ERROR))
                span.record_exception(e)
                span.end()
                raise
            output_formatter = to_wrap.get("output_formatter") or (
                lambda x: json_dumps(x)
            )
            span.set_attribute("lmnr.span.output", output_formatter(result))
            return result


@_with_wrapper
async def _wrap_async(
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    if to_wrap.get("action") == "start_parent_span":
        parent_span = Laminar.start_span("computer.run")
        add_input_to_parent_span(parent_span, instance)
        result = await wrapped(*args, **kwargs)
        try:
            instance._interface._lmnr_parent_span = parent_span
        except Exception:
            pass
        return result
    elif to_wrap.get("action") == "end_parent_span":
        result = await wrapped(*args, **kwargs)
        try:
            parent_span: Span = instance._interface._lmnr_parent_span
            if parent_span and parent_span.is_recording():
                parent_span.end()
        except Exception:
            pass
        return result

    # if there's no parent span, use
    parent_span = trace.get_current_span(context=get_current_context())
    try:
        parent_span: Span = instance._lmnr_parent_span
    except Exception:
        pass

    with Laminar.use_span(parent_span):
        instance_name = "interface"
        with Laminar.start_as_current_span(
            f"{instance_name}.{to_wrap.get('method')}",
            span_type="TOOL",
        ) as span:
            span.set_attribute(
                "lmnr.span.input",
                json_dumps(get_input_from_func_args(wrapped, True, args, kwargs)),
            )
            try:
                result = await wrapped(*args, **kwargs)
            except Exception as e:  # pylint: disable=broad-except
                span.set_status(Status(StatusCode.ERROR))
                span.record_exception(e)
                span.end()
                raise
            output_formatter = to_wrap.get("output_formatter") or (
                lambda x: json_dumps(x)
            )
            span.set_attribute("lmnr.span.output", output_formatter(result))
            return result


class CuaComputerInstrumentor(BaseInstrumentor):
    def __init__(self):
        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")

            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _wrap(wrapped_method),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _wrap_async(wrapped_method),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            try:
                unwrap(
                    f"{wrap_package}.{wrap_object}",
                    wrapped_method.get("method"),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            try:
                unwrap(
                    f"{wrap_package}.{wrap_object}",
                    wrapped_method.get("method"),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist



import asyncio
import os
import time

import orjson

from opentelemetry import trace

from lmnr.opentelemetry_lib.tracing.context import get_current_context
from lmnr.opentelemetry_lib.tracing import TracerWrapper
from lmnr.opentelemetry_lib.utils.package_check import is_package_installed
from lmnr.sdk.decorators import observe
from lmnr.sdk.browser.utils import retry_sync, retry_async
from lmnr.sdk.browser.background_send_events import (
    get_background_loop,
    track_async_send,
)
from lmnr.sdk.client.asynchronous.async_client import AsyncLaminarClient
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.types import MaskInputOptions

try:
    if is_package_installed("playwright"):
        from playwright.async_api import Page
        from playwright.sync_api import Page as SyncPage
    elif is_package_installed("patchright"):
        from patchright.async_api import Page
        from patchright.sync_api import Page as SyncPage
    else:
        raise ImportError(
            "Attempted to import lmnr.sdk.browser.pw_utils, but neither "
            "playwright nor patchright is installed. Use `pip install playwright` "
            "or `pip install patchright` to install one of the supported browsers."
        )
except ImportError as e:
    raise ImportError(
        "Attempted to import lmnr.sdk.browser.pw_utils, but neither "
        "playwright nor patchright is installed. Use `pip install playwright` "
        "or `pip install patchright` to install one of the supported browsers."
    ) from e

logger = get_default_logger(__name__)

OLD_BUFFER_TIMEOUT = 60


def create_send_events_handler(
    chunk_buffers: dict,
    session_id: str,
    trace_id: str,
    client: AsyncLaminarClient,
    background_loop: asyncio.AbstractEventLoop,
):
    """
    Create an async event handler for sending browser events.

    This handler reassembles chunked event data and submits it to the background
    loop for async HTTP sending. The handler itself processes chunks synchronously
    but delegates the actual HTTP send to the background loop.

    Args:
        chunk_buffers: Dictionary to store incomplete chunk batches
        session_id: Browser session ID
        trace_id: OpenTelemetry trace ID
        client: Async Laminar client for HTTP requests
        background_loop: Background event loop for async sends

    Returns:
        An async function that handles incoming event chunks from the browser
    """

    async def send_events_from_browser(chunk):
        try:
            # Handle chunked data
            batch_id = chunk["batchId"]
            chunk_index = chunk["chunkIndex"]
            total_chunks = chunk["totalChunks"]
            data = chunk["data"]

            # Initialize buffer for this batch if needed
            if batch_id not in chunk_buffers:
                chunk_buffers[batch_id] = {
                    "chunks": {},
                    "total": total_chunks,
                    "timestamp": time.time(),
                }

            # Store chunk
            chunk_buffers[batch_id]["chunks"][chunk_index] = data

            # Check if we have all chunks
            if len(chunk_buffers[batch_id]["chunks"]) == total_chunks:
                # Reassemble the full message
                full_data = ""
                for i in range(total_chunks):
                    full_data += chunk_buffers[batch_id]["chunks"][i]

                # Parse the JSON
                events = orjson.loads(full_data)

                # Send to server in background loop (independent of Playwright's loop)
                if events and len(events) > 0:
                    future = asyncio.run_coroutine_threadsafe(
                        client._browser_events.send(session_id, trace_id, events),
                        background_loop,
                    )
                    track_async_send(future)

                # Clean up buffer
                del chunk_buffers[batch_id]

            # Clean up old incomplete buffers
            current_time = time.time()
            to_delete = []
            for bid, buffer in chunk_buffers.items():
                if current_time - buffer["timestamp"] > OLD_BUFFER_TIMEOUT:
                    to_delete.append(bid)
            for bid in to_delete:
                logger.debug(f"Cleaning up incomplete chunk buffer: {bid}")
                del chunk_buffers[bid]

        except Exception as e:
            logger.debug(f"Could not send events: {e}")

    return send_events_from_browser


current_dir = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(current_dir, "recorder", "record.umd.min.cjs"), "r") as f:
    RRWEB_CONTENT = f"() => {{ {f.read()} }}"

with open(os.path.join(current_dir, "inject_script.js"), "r") as f:
    INJECT_SCRIPT_CONTENT = f.read()


def get_mask_input_setting() -> MaskInputOptions:
    """Get the mask_input setting from session recording configuration."""
    try:
        config = TracerWrapper.get_session_recording_options()
        return config.get(
            "mask_input_options",
            MaskInputOptions(
                textarea=False,
                text=False,
                number=False,
                select=False,
                email=False,
                tel=False,
            ),
        )
    except (AttributeError, Exception):
        # Fallback to default configuration if TracerWrapper is not initialized
        return MaskInputOptions(
            textarea=False,
            text=False,
            number=False,
            select=False,
            email=False,
            tel=False,
        )


def inject_session_recorder_sync(page: SyncPage):
    try:
        try:
            is_loaded = page.evaluate(
                """() => typeof window.lmnrRrweb !== 'undefined'"""
            )
        except Exception as e:
            logger.debug(f"Failed to check if session recorder is loaded: {e}")
            is_loaded = False

        if not is_loaded:

            def load_session_recorder():
                try:
                    if page.is_closed():
                        return False
                    page.evaluate(RRWEB_CONTENT)
                    return True
                except Exception as e:
                    logger.debug(f"Failed to load session recorder: {e}")
                    return False

            if not retry_sync(
                load_session_recorder,
                delay=1,
                error_message="Failed to load session recorder",
            ):
                return

            try:
                if not page.is_closed():
                    page.evaluate(
                        f"({INJECT_SCRIPT_CONTENT})({orjson.dumps(get_mask_input_setting()).decode('utf-8')}, false)"
                    )
            except Exception as e:
                logger.debug(f"Failed to inject session recorder: {e}")

    except Exception as e:
        logger.error(f"Error during session recorder injection: {e}")


async def inject_session_recorder_async(page: Page):
    try:
        try:
            is_loaded = await page.evaluate(
                """() => typeof window.lmnrRrweb !== 'undefined'"""
            )
        except Exception as e:
            logger.debug(f"Failed to check if session recorder is loaded: {e}")
            is_loaded = False

        if not is_loaded:

            async def load_session_recorder():
                try:
                    if page.is_closed():
                        return False
                    await page.evaluate(RRWEB_CONTENT)
                    return True
                except Exception as e:
                    logger.debug(f"Failed to load session recorder: {e}")
                    return False

            if not await retry_async(
                load_session_recorder,
                delay=1,
                error_message="Failed to load session recorder",
            ):
                return

            try:
                if not page.is_closed():
                    await page.evaluate(
                        f"({INJECT_SCRIPT_CONTENT})({orjson.dumps(get_mask_input_setting()).decode('utf-8')}, false)"
                    )
            except Exception as e:
                logger.debug(f"Failed to inject session recorder placeholder: {e}")

    except Exception as e:
        logger.error(f"Error during session recorder injection: {e}")


@observe(name="playwright.page", ignore_input=True, ignore_output=True)
def start_recording_events_sync(
    page: SyncPage, session_id: str, client: AsyncLaminarClient
):

    ctx = get_current_context()
    span = trace.get_current_span(ctx)
    trace_id = format(span.get_span_context().trace_id, "032x")
    span.set_attribute("lmnr.internal.has_browser_session", True)

    # Get the background loop for async sends
    background_loop = get_background_loop()

    # Buffer for reassembling chunks
    chunk_buffers = {}

    # Create the async event handler (shared implementation)
    send_events_from_browser = create_send_events_handler(
        chunk_buffers, session_id, trace_id, client, background_loop
    )

    def submit_event(chunk):
        """Sync wrapper that submits async handler to background loop."""
        try:
            # Submit async handler to background loop
            asyncio.run_coroutine_threadsafe(
                send_events_from_browser(chunk),
                background_loop,
            )
        except Exception as e:
            logger.debug(f"Error submitting event: {e}")

    try:
        page.expose_function("lmnrSendEvents", submit_event)
    except Exception as e:
        logger.debug(f"Could not expose function: {e}")

    inject_session_recorder_sync(page)

    def on_load(p):
        try:
            if not p.is_closed():
                inject_session_recorder_sync(p)
        except Exception as e:
            logger.debug(f"Error in on_load handler: {e}")

    page.on("domcontentloaded", on_load)


@observe(name="playwright.page", ignore_input=True, ignore_output=True)
async def start_recording_events_async(
    page: Page, session_id: str, client: AsyncLaminarClient
):
    ctx = get_current_context()
    span = trace.get_current_span(ctx)
    trace_id = format(span.get_span_context().trace_id, "032x")
    span.set_attribute("lmnr.internal.has_browser_session", True)

    # Get the background loop for async sends (independent of Playwright's loop)
    background_loop = get_background_loop()

    # Buffer for reassembling chunks
    chunk_buffers = {}

    # Create the async event handler (shared implementation)
    send_events_from_browser = create_send_events_handler(
        chunk_buffers, session_id, trace_id, client, background_loop
    )

    try:
        await page.expose_function("lmnrSendEvents", send_events_from_browser)
    except Exception as e:
        logger.debug(f"Could not expose function: {e}")

    await inject_session_recorder_async(page)

    async def on_load(p):
        try:
            # Check if page is closed before attempting to inject
            if not p.is_closed():
                await inject_session_recorder_async(p)
        except Exception as e:
            logger.debug(f"Error in on_load handler: {e}")

    page.on("domcontentloaded", on_load)


def take_full_snapshot(page: Page):
    return page.evaluate(
        """() => {
        if (window.lmnrRrweb) {
            try {
                window.lmnrRrweb.record.takeFullSnapshot();
                return true;
            } catch (e) {
                console.error("Error taking full snapshot:", e);
                return false;
            }
        }
        return false;
    }"""
    )


async def take_full_snapshot_async(page: Page):
    return await page.evaluate(
        """() => {
        if (window.lmnrRrweb) {
            try {
                window.lmnrRrweb.record.takeFullSnapshot();
                return true;
            } catch (e) {
                console.error("Error taking full snapshot:", e);
                return false;
            }
        }
        return false;
    }"""
    )



"""Evals resource for interacting with Laminar evaluations API."""

import uuid
import warnings

from typing import Any

from lmnr.sdk.client.synchronous.resources.base import BaseResource
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.types import (
    GetDatapointsResponse,
    EvaluationResultDatapoint,
    InitEvaluationResponse,
    PartialEvaluationDatapoint,
)
from lmnr.sdk.utils import serialize

INITIAL_EVALUATION_DATAPOINT_MAX_DATA_LENGTH = 16_000_000  # 16MB
logger = get_default_logger(__name__)


class Evals(BaseResource):
    """Resource for interacting with Laminar evaluations API."""

    def init(
        self,
        name: str | None = None,
        group_name: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> InitEvaluationResponse:
        """Initialize a new evaluation.

        Args:
            name (str | None, optional): Name of the evaluation. Defaults to None.
            group_name (str | None, optional): Group name for the evaluation. Defaults to None.
            metadata (dict[str, Any] | None, optional): Metadata to associate with. Defaults to None.

        Returns:
            InitEvaluationResponse: The response from the initialization request.
        """
        response = self._client.post(
            self._base_url + "/v1/evals",
            json={
                "name": name,
                "groupName": group_name,
                "metadata": metadata,
            },
            headers=self._headers(),
        )
        if response.status_code != 200:
            if response.status_code == 401:
                raise ValueError("Unauthorized. Please check your project API key.")
            raise ValueError(f"Error initializing evaluation: {response.text}")
        resp_json = response.json()
        return InitEvaluationResponse.model_validate(resp_json)

    def create_evaluation(
        self,
        name: str | None = None,
        group_name: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> uuid.UUID:
        """
        Create a new evaluation and return its ID.

        Parameters:
            name (str | None, optional): Optional name of the evaluation.
            group_name (str | None, optional): An identifier to group evaluations.
            metadata (dict[str, Any] | None, optional): Metadata to associate with. Defaults to None.

        Returns:
            uuid.UUID: The evaluation ID.
        """
        evaluation = self.init(name=name, group_name=group_name, metadata=metadata)
        return evaluation.id

    def create_datapoint(
        self,
        eval_id: uuid.UUID,
        data: Any,
        target: Any = None,
        metadata: dict[str, Any] | None = None,
        index: int | None = None,
        trace_id: uuid.UUID | None = None,
    ) -> uuid.UUID:
        """
        Create a datapoint for an evaluation.

        Parameters:
            eval_id (uuid.UUID): The evaluation ID.
            data: The input data for the executor.
            target: The target/expected output for evaluators.
            metadata (dict[str, Any] | None, optional): Optional metadata.
            index (int | None, optional): Optional index of the datapoint.
            trace_id (uuid.UUID | None, optional): Optional trace ID.

        Returns:
            uuid.UUID: The datapoint ID.
        """

        datapoint_id = uuid.uuid4()

        # Create a minimal datapoint first
        partial_datapoint = PartialEvaluationDatapoint(
            id=datapoint_id,
            data=data,
            target=target,
            index=index or 0,
            trace_id=trace_id or uuid.uuid4(),
            executor_span_id=uuid.uuid4(),  # Will be updated when executor runs
            metadata=metadata,
        )

        self.save_datapoints(eval_id, [partial_datapoint])
        return datapoint_id

    def save_datapoints(
        self,
        eval_id: uuid.UUID,
        datapoints: list[EvaluationResultDatapoint | PartialEvaluationDatapoint],
        group_name: str | None = None,
    ):
        """Save evaluation datapoints.

        Args:
            eval_id (uuid.UUID): The evaluation ID.
            datapoints (list[EvaluationResultDatapoint | PartialEvaluationDatapoint]): The datapoints to save.
            group_name (str | None, optional): Group name for the datapoints. Defaults to None.

        Raises:
            ValueError: If there's an error saving the datapoints.
        """
        length = INITIAL_EVALUATION_DATAPOINT_MAX_DATA_LENGTH
        points = [datapoint.to_dict(max_data_length=length) for datapoint in datapoints]
        response = self._client.post(
            self._base_url + f"/v1/evals/{eval_id}/datapoints",
            json={
                "points": points,
                "groupName": group_name,
            },
            headers=self._headers(),
        )
        if response.status_code == 413:
            self._retry_save_datapoints(eval_id, datapoints, group_name)
            return

        if response.status_code != 200:
            raise ValueError(
                f"Error saving evaluation datapoints: [{response.status_code}] {response.text}"
            )

    def update_datapoint(
        self,
        eval_id: uuid.UUID,
        datapoint_id: uuid.UUID,
        scores: dict[str, float | int],
        executor_output: Any | None = None,
    ) -> None:
        """Update a datapoint with evaluation results.

        Args:
            eval_id (uuid.UUID): The evaluation ID.
            datapoint_id (uuid.UUID): The datapoint ID.
            executor_output (Any): The executor output.
            scores (dict[str, float | int] | None, optional): The scores. Defaults to None.
        """

        response = self._client.post(
            self._base_url + f"/v1/evals/{eval_id}/datapoints/{datapoint_id}",
            json={
                "executorOutput": (
                    str(serialize(executor_output))[
                        :INITIAL_EVALUATION_DATAPOINT_MAX_DATA_LENGTH
                    ]
                    if executor_output is not None
                    else None
                ),
                "scores": scores,
            },
            headers=self._headers(),
        )

        if response.status_code != 200:
            raise ValueError(f"Error updating evaluation datapoint: {response.text}")

    def get_datapoints(
        self,
        dataset_name: str,
        offset: int,
        limit: int,
    ) -> GetDatapointsResponse:
        """Get datapoints from a dataset.

        Args:
            dataset_name (str): The name of the dataset.
            offset (int): The offset to start from.
            limit (int): The maximum number of datapoints to return.

        Returns:
            GetDatapointsResponse: The response containing the datapoints.

        Raises:
            ValueError: If there's an error fetching the datapoints.
        """

        warnings.warn(
            "Use client.datasets.pull instead",
            DeprecationWarning,
        )

        params = {"name": dataset_name, "offset": offset, "limit": limit}
        response = self._client.get(
            self._base_url + "/v1/datasets/datapoints",
            params=params,
            headers=self._headers(),
        )
        if response.status_code != 200:
            try:
                resp_json = response.json()
                raise ValueError(
                    f"Error fetching datapoints: [{response.status_code}] {resp_json}"
                )
            except Exception:
                raise ValueError(
                    f"Error fetching datapoints: [{response.status_code}] {response.text}"
                )
        return GetDatapointsResponse.model_validate(response.json())

    def _retry_save_datapoints(
        self,
        eval_id: uuid.UUID,
        datapoints: list[EvaluationResultDatapoint | PartialEvaluationDatapoint],
        group_name: str | None = None,
        initial_length: int = INITIAL_EVALUATION_DATAPOINT_MAX_DATA_LENGTH,
        max_retries: int = 20,
    ):
        retry = 0
        length = initial_length
        while retry < max_retries:
            retry += 1
            length = length // 2
            logger.debug(
                f"Retrying save datapoints: {retry} of {max_retries}, length: {length}"
            )
            if length == 0:
                raise ValueError("Error saving evaluation datapoints")
            points = [
                datapoint.to_dict(max_data_length=length) for datapoint in datapoints
            ]
            response = self._client.post(
                self._base_url + f"/v1/evals/{eval_id}/datapoints",
                json={
                    "points": points,
                    "groupName": group_name,
                },
                headers=self._headers(),
            )
            if response.status_code != 413:
                break
        if response.status_code != 200:
            raise ValueError(
                f"Error saving evaluation datapoints: [{response.status_code}] {response.text}"
            )



import logging

from opentelemetry import context as context_api
from ..shared import (
    _set_client_attributes,
    _set_functions_attributes,
    _set_request_attributes,
    _set_response_attributes,
    _set_span_attribute,
    _set_span_stream_usage,
    get_token_count_from_string,
    is_streaming_response,
    model_as_dict,
    propagate_trace_context,
    should_record_stream_token_usage,
)
from ..shared.config import Config
from ..shared.event_emitter import emit_event
from ..shared.event_models import (
    ChoiceEvent,
    MessageEvent,
)
from ..utils import (
    _with_tracer_wrapper,
    dont_throw,
    is_openai_v1,
    should_emit_events,
    should_send_prompts,
)
from lmnr.opentelemetry_lib.tracing.context import (
    get_current_context,
    get_event_attributes_from_context,
)
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv_ai import (
    SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY,
    LLMRequestTypeValues,
    SpanAttributes,
)
from opentelemetry.trace import SpanKind
from opentelemetry.trace.status import Status, StatusCode

SPAN_NAME = "openai.completion"
LLM_REQUEST_TYPE = LLMRequestTypeValues.COMPLETION

logger = logging.getLogger(__name__)


@_with_tracer_wrapper
def completion_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return wrapped(*args, **kwargs)

    # span needs to be opened and closed manually because the response is a generator
    span = tracer.start_span(
        SPAN_NAME,
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLM_REQUEST_TYPE.value},
        context=get_current_context(),
    )

    _handle_request(span, kwargs, instance)

    try:
        response = wrapped(*args, **kwargs)
    except Exception as e:
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        attributes = get_event_attributes_from_context()
        span.record_exception(e, attributes=attributes)
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.end()
        raise

    if is_streaming_response(response):
        # span will be closed after the generator is done
        return _build_from_streaming_response(span, kwargs, response)
    else:
        _handle_response(response, span, instance)

    span.end()
    return response


@_with_tracer_wrapper
async def acompletion_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return await wrapped(*args, **kwargs)

    span = tracer.start_span(
        name=SPAN_NAME,
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLM_REQUEST_TYPE.value},
        context=get_current_context(),
    )

    _handle_request(span, kwargs, instance)

    try:
        response = await wrapped(*args, **kwargs)
    except Exception as e:
        span.set_attribute(ERROR_TYPE, e.__class__.__name__)
        attributes = get_event_attributes_from_context()
        span.record_exception(e, attributes=attributes)
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.end()
        raise

    if is_streaming_response(response):
        # span will be closed after the generator is done
        return _abuild_from_streaming_response(span, kwargs, response)
    else:
        _handle_response(response, span, instance)

    span.end()
    return response


@dont_throw
def _handle_request(span, kwargs, instance):
    _set_request_attributes(span, kwargs, instance)
    if should_emit_events():
        _emit_prompts_events(kwargs)
    else:
        if should_send_prompts():
            _set_prompts(span, kwargs.get("prompt"))
            _set_functions_attributes(span, kwargs.get("functions"))
    _set_client_attributes(span, instance)
    if Config.enable_trace_context_propagation:
        propagate_trace_context(span, kwargs)


def _emit_prompts_events(kwargs):
    prompt = kwargs.get("prompt")
    if isinstance(prompt, list):
        for p in prompt:
            emit_event(MessageEvent(content=p))
    elif isinstance(prompt, str):
        emit_event(MessageEvent(content=prompt))


@dont_throw
def _handle_response(response, span, instance=None):
    if is_openai_v1():
        response_dict = model_as_dict(response)
    else:
        response_dict = response

    _set_response_attributes(span, response_dict)
    if should_emit_events():
        for choice in response.choices:
            emit_event(_parse_choice_event(choice))
    else:
        if should_send_prompts():
            _set_completions(span, response_dict.get("choices"))


def _set_prompts(span, prompt):
    if not span.is_recording() or not prompt:
        return

    _set_span_attribute(
        span,
        f"{SpanAttributes.LLM_PROMPTS}.0.user",
        prompt[0] if isinstance(prompt, list) else prompt,
    )


@dont_throw
def _set_completions(span, choices):
    if not span.is_recording() or not choices:
        return

    for choice in choices:
        index = choice.get("index")
        prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
        _set_span_attribute(
            span, f"{prefix}.finish_reason", choice.get("finish_reason")
        )
        _set_span_attribute(span, f"{prefix}.content", choice.get("text"))


@dont_throw
def _build_from_streaming_response(span, request_kwargs, response):
    complete_response = {"choices": [], "model": "", "id": ""}
    for item in response:
        yield item
        _accumulate_streaming_response(complete_response, item)

    _set_response_attributes(span, complete_response)

    _set_token_usage(span, request_kwargs, complete_response)

    if should_emit_events():
        _emit_streaming_response_events(complete_response)
    else:
        if should_send_prompts():
            _set_completions(span, complete_response.get("choices"))

    span.set_status(Status(StatusCode.OK))
    span.end()


@dont_throw
async def _abuild_from_streaming_response(span, request_kwargs, response):
    complete_response = {"choices": [], "model": "", "id": ""}
    async for item in response:
        yield item
        _accumulate_streaming_response(complete_response, item)

    _set_response_attributes(span, complete_response)

    _set_token_usage(span, request_kwargs, complete_response)

    if should_emit_events():
        _emit_streaming_response_events(complete_response)
    else:
        if should_send_prompts():
            _set_completions(span, complete_response.get("choices"))

    span.set_status(Status(StatusCode.OK))
    span.end()


def _emit_streaming_response_events(complete_response):
    for i, choice in enumerate(complete_response["choices"]):
        emit_event(
            ChoiceEvent(
                index=choice.get("index", i),
                message={"content": choice.get("text"), "role": "assistant"},
                finish_reason=choice.get("finish_reason", "unknown"),
            )
        )


@dont_throw
def _set_token_usage(span, request_kwargs, complete_response):
    # use tiktoken calculate token usage
    if should_record_stream_token_usage():
        prompt_usage = -1
        completion_usage = -1

        # prompt_usage
        if request_kwargs and request_kwargs.get("prompt"):
            prompt_content = request_kwargs.get("prompt")
            model_name = complete_response.get("model") or None

            if model_name:
                prompt_usage = get_token_count_from_string(prompt_content, model_name)

        # completion_usage
        if complete_response.get("choices"):
            completion_content = ""
            model_name = complete_response.get("model") or None

            for choice in complete_response.get("choices"):
                if choice.get("text"):
                    completion_content += choice.get("text")

            if model_name:
                completion_usage = get_token_count_from_string(
                    completion_content, model_name
                )

        # span record
        _set_span_stream_usage(span, prompt_usage, completion_usage)


@dont_throw
def _accumulate_streaming_response(complete_response, item):
    if is_openai_v1():
        item = model_as_dict(item)

    complete_response["model"] = item.get("model")
    complete_response["id"] = item.get("id")
    for choice in item.get("choices"):
        index = choice.get("index")
        if len(complete_response.get("choices")) <= index:
            complete_response["choices"].append({"index": index, "text": ""})
        complete_choice = complete_response.get("choices")[index]
        if choice.get("finish_reason"):
            complete_choice["finish_reason"] = choice.get("finish_reason")

        if choice.get("text"):
            complete_choice["text"] += choice.get("text")

    return complete_response


def _parse_choice_event(choice) -> ChoiceEvent:
    has_message = choice.text is not None
    has_finish_reason = choice.finish_reason is not None

    content = choice.text if has_message else None
    finish_reason = choice.finish_reason if has_finish_reason else "unknown"

    return ChoiceEvent(
        index=choice.index,
        message={"content": content, "role": "assistant"},
        finish_reason=finish_reason,
    )



from pathlib import Path
from typing import Any
import csv
import orjson

from lmnr.sdk.log import get_default_logger

LOG = get_default_logger(__name__, verbose=False)


def _is_supported_file(file: Path) -> bool:
    """Check if a file is supported."""
    return file.suffix in [".json", ".csv", ".jsonl"]


def _collect_files(paths: list[Path], recursive: bool = False) -> list[Path]:
    """
    Collect all supported files from the given paths.

    Handles both files and directories. If a path is a directory,
    collects all supported files within it (recursively if specified).
    """
    collected_files = []

    for path in paths:
        if path.is_file():
            if _is_supported_file(path):
                collected_files.append(path)
            else:
                LOG.warning(f"Skipping unsupported file type: {path}")
        elif path.is_dir():
            for item in path.iterdir():
                if item.is_file() and _is_supported_file(item):
                    collected_files.append(item)
                elif recursive and item.is_dir():
                    # Recursively collect files from subdirectories
                    collected_files.extend(_collect_files([item], recursive=True))
        else:
            LOG.warning(f"Path does not exist or is not accessible: {path}")

    return collected_files


def _read_file(file: Path) -> list[dict[str, Any]]:
    """Read data from a single file and return as a list of dictionaries."""
    if file.suffix == ".json":
        result = orjson.loads(file.read_bytes())
        if isinstance(result, list):
            return result
        else:
            return [result]
    elif file.suffix == ".csv":
        return [dict(row) for row in csv.DictReader(file.read_text().splitlines())]
    elif file.suffix == ".jsonl":
        return [
            orjson.loads(line) for line in file.read_text().splitlines() if line.strip()
        ]
    else:
        raise ValueError(f"Unsupported file type: {file.suffix}")


def load_from_paths(paths: list[Path], recursive: bool = False) -> list[dict[str, Any]]:
    """
    Load data from all files in the specified paths.

    First collects all file paths, then reads each file's data.
    """
    files = _collect_files(paths, recursive)

    if not files:
        LOG.warning("No supported files found in the specified paths")
        return []

    LOG.info(f"Found {len(files)} file(s) to read")

    result = []
    for file in files:
        try:
            data = _read_file(file)
            result.extend(data)
            LOG.info(f"Read {len(data)} record(s) from {file}")
        except Exception as e:
            LOG.error(f"Error reading file {file}: {e}")
            raise

    return result


def parse_paths(paths: list[str]) -> list[Path]:
    """Parse paths."""
    return [Path(path) for path in paths]



from lmnr.sdk.client.synchronous.resources.browser_events import BrowserEvents
from lmnr.sdk.client.synchronous.resources.evals import Evals
from lmnr.sdk.client.synchronous.resources.tags import Tags
from lmnr.sdk.client.synchronous.resources.evaluators import Evaluators

__all__ = ["Evals", "Evaluators", "BrowserEvents", "Tags"]



from __future__ import annotations  # For "Self" | str | ... type hint

import json
import logging
import datetime
from pydantic import BaseModel, Field
import uuid

from enum import Enum
from opentelemetry.trace import SpanContext, TraceFlags
from typing import Any, Awaitable, Callable, Optional
from typing_extensions import TypedDict  # compatibility with python < 3.12

from .utils import serialize, json_dumps

DEFAULT_DATAPOINT_MAX_DATA_LENGTH = 16_000_000  # 16MB


Numeric = int | float
NumericTypes = (int, float)  # for use with isinstance

EvaluationDatapointData = Any  # non-null, must be JSON-serializable
EvaluationDatapointTarget = Any | None  # must be JSON-serializable
EvaluationDatapointMetadata = Any | None  # must be JSON-serializable


# EvaluationDatapoint is a single data point in the evaluation
class Datapoint(BaseModel):
    # input to the executor function.
    data: EvaluationDatapointData
    # input to the evaluator function (alongside the executor output).
    target: EvaluationDatapointTarget = Field(default_factory=dict)
    metadata: EvaluationDatapointMetadata = Field(default_factory=dict)
    id: uuid.UUID | None = Field(default=None)
    created_at: datetime.datetime | None = Field(default=None, alias="createdAt")


class Dataset(BaseModel):
    id: uuid.UUID = Field()
    name: str = Field()
    created_at: datetime.datetime = Field(alias="createdAt")


class PushDatapointsResponse(BaseModel):
    dataset_id: uuid.UUID = Field(alias="datasetId")


ExecutorFunctionReturnType = Any
EvaluatorFunctionReturnType = Numeric | dict[str, Numeric]

ExecutorFunction = Callable[
    [EvaluationDatapointData, Any],
    ExecutorFunctionReturnType | Awaitable[ExecutorFunctionReturnType],
]

# EvaluatorFunction is a function that takes the output of the executor and the
# target data, and returns a score. The score can be a single number or a
# record of string keys and number values. The latter is useful for evaluating
# multiple criteria in one go instead of running multiple evaluators.
EvaluatorFunction = Callable[
    [ExecutorFunctionReturnType, Any],
    EvaluatorFunctionReturnType | Awaitable[EvaluatorFunctionReturnType],
]


class HumanEvaluatorOptionsEntry(TypedDict):
    label: str
    value: float


class HumanEvaluator(BaseModel):
    options: list[HumanEvaluatorOptionsEntry] = Field(default_factory=list)


class InitEvaluationResponse(BaseModel):
    id: uuid.UUID
    createdAt: datetime.datetime
    groupId: str
    name: str
    projectId: uuid.UUID


class EvaluationDatapointDatasetLink(BaseModel):
    dataset_id: uuid.UUID
    datapoint_id: uuid.UUID
    created_at: datetime.datetime

    def to_dict(self):
        return {
            "datasetId": str(self.dataset_id),
            "datapointId": str(self.datapoint_id),
            "createdAt": self.created_at.isoformat(),
        }


class PartialEvaluationDatapoint(BaseModel):
    id: uuid.UUID
    data: EvaluationDatapointData
    target: EvaluationDatapointTarget
    index: int
    trace_id: uuid.UUID
    executor_span_id: uuid.UUID
    metadata: EvaluationDatapointMetadata = Field(default=None)
    dataset_link: EvaluationDatapointDatasetLink | None = Field(default=None)

    # uuid is not serializable by default, so we need to convert it to a string
    def to_dict(self, max_data_length: int = DEFAULT_DATAPOINT_MAX_DATA_LENGTH):
        serialized_data = serialize(self.data)
        serialized_target = serialize(self.target)
        str_data = json_dumps(serialized_data)
        str_target = json_dumps(serialized_target)
        try:
            return {
                "id": str(self.id),
                "data": (
                    str_data[:max_data_length]
                    if len(str_data) > max_data_length
                    else serialized_data
                ),
                "target": (
                    str_target[:max_data_length]
                    if len(str_target) > max_data_length
                    else serialized_target
                ),
                "index": self.index,
                "traceId": str(self.trace_id),
                "executorSpanId": str(self.executor_span_id),
                "metadata": (
                    serialize(self.metadata) if self.metadata is not None else {}
                ),
                "datasetLink": (
                    self.dataset_link.to_dict()
                    if self.dataset_link is not None
                    else None
                ),
            }
        except Exception as e:
            raise ValueError(f"Error serializing PartialEvaluationDatapoint: {e}")


class EvaluationResultDatapoint(BaseModel):
    id: uuid.UUID
    index: int
    data: EvaluationDatapointData
    target: EvaluationDatapointTarget
    executor_output: ExecutorFunctionReturnType
    scores: dict[str, Optional[Numeric]]
    trace_id: uuid.UUID
    executor_span_id: uuid.UUID
    metadata: EvaluationDatapointMetadata = Field(default=None)
    dataset_link: EvaluationDatapointDatasetLink | None = Field(default=None)

    # uuid is not serializable by default, so we need to convert it to a string
    def to_dict(self, max_data_length: int = DEFAULT_DATAPOINT_MAX_DATA_LENGTH):
        try:
            serialized_data = serialize(self.data)
            serialized_target = serialize(self.target)
            serialized_executor_output = serialize(self.executor_output)
            str_data = json.dumps(serialized_data)
            str_target = json.dumps(serialized_target)
            str_executor_output = json.dumps(serialized_executor_output)
            return {
                # preserve only preview of the data, target and executor output
                # (full data is in trace)
                "id": str(self.id),
                "data": (
                    str_data[:max_data_length]
                    if len(str_data) > max_data_length
                    else serialized_data
                ),
                "target": (
                    str_target[:max_data_length]
                    if len(str_target) > max_data_length
                    else serialized_target
                ),
                "executorOutput": (
                    str_executor_output[:max_data_length]
                    if len(str_executor_output) > max_data_length
                    else serialized_executor_output
                ),
                "scores": self.scores,
                "traceId": str(self.trace_id),
                "executorSpanId": str(self.executor_span_id),
                "index": self.index,
                "metadata": (
                    serialize(self.metadata) if self.metadata is not None else {}
                ),
                "datasetLink": (
                    self.dataset_link.to_dict()
                    if self.dataset_link is not None
                    else None
                ),
            }
        except Exception as e:
            raise ValueError(f"Error serializing EvaluationResultDatapoint: {e}")


class SpanType(Enum):
    DEFAULT = "DEFAULT"
    LLM = "LLM"
    PIPELINE = "PIPELINE"  # must not be set manually
    EXECUTOR = "EXECUTOR"
    EVALUATOR = "EVALUATOR"
    HUMAN_EVALUATOR = "HUMAN_EVALUATOR"
    EVALUATION = "EVALUATION"


class TraceType(Enum):
    DEFAULT = "DEFAULT"
    EVALUATION = "EVALUATION"


class GetDatapointsResponse(BaseModel):
    items: list[Datapoint]
    total_count: int = Field(alias="totalCount")


class LaminarSpanContext(BaseModel):
    """
    A span context that can be used to continue a trace across services. This
    is a slightly modified version of the OpenTelemetry span context. For
    usage examples, see `Laminar.serialize_span_context`,
    `Laminar.get_span_context`, and `Laminar.deserialize_laminar_span_context`.

    The difference between this and the OpenTelemetry span context is that
    the `trace_id` and `span_id` are stored as UUIDs instead of integers for
    easier debugging, and the separate trace flags are not currently stored.
    """

    trace_id: uuid.UUID
    span_id: uuid.UUID
    is_remote: bool = Field(default=False)
    span_path: list[str] = Field(default=[])
    span_ids_path: list[str] = Field(default=[])  # stringified UUIDs
    user_id: str | None = Field(default=None)
    session_id: str | None = Field(default=None)
    trace_type: TraceType | None = Field(default=None)
    metadata: dict[str, Any] | None = Field(default=None)

    def __str__(self) -> str:
        return self.model_dump_json()

    @classmethod
    def try_to_otel_span_context(
        cls,
        span_context: "LaminarSpanContext" | dict[str, Any] | str | SpanContext,
        logger: logging.Logger | None = None,
    ) -> SpanContext:
        if logger is None:
            logger = logging.getLogger(__name__)

        if isinstance(span_context, LaminarSpanContext):
            return SpanContext(
                trace_id=span_context.trace_id.int,
                span_id=span_context.span_id.int,
                is_remote=span_context.is_remote,
                trace_flags=TraceFlags(TraceFlags.SAMPLED),
            )
        elif isinstance(span_context, SpanContext) or (
            isinstance(getattr(span_context, "trace_id", None), int)
            and isinstance(getattr(span_context, "span_id", None), int)
        ):
            logger.warning(
                "span_context provided"
                " is likely a raw OpenTelemetry span context. Will try to use it. "
                "Please use `LaminarSpanContext` instead."
            )
            return span_context
        elif isinstance(span_context, (dict, str)):
            try:
                laminar_span_context = cls.deserialize(span_context)
                return SpanContext(
                    trace_id=laminar_span_context.trace_id.int,
                    span_id=laminar_span_context.span_id.int,
                    is_remote=laminar_span_context.is_remote,
                    trace_flags=TraceFlags(TraceFlags.SAMPLED),
                )
            except Exception:
                raise ValueError("Invalid span_context provided")
        else:
            raise ValueError("Invalid span_context provided")

    @classmethod
    def deserialize(cls, data: dict[str, Any] | str) -> "LaminarSpanContext":
        if isinstance(data, dict):
            # Convert camelCase to snake_case for known fields
            converted_data = {
                "trace_id": data.get("trace_id") or data.get("traceId"),
                "span_id": data.get("span_id") or data.get("spanId"),
                "is_remote": data.get("is_remote") or data.get("isRemote", False),
                "span_path": data.get("span_path") or data.get("spanPath", []),
                "span_ids_path": data.get("span_ids_path")
                or data.get("spanIdsPath", []),
                "user_id": data.get("user_id") or data.get("userId"),
                "session_id": data.get("session_id") or data.get("sessionId"),
                "trace_type": data.get("trace_type") or data.get("traceType"),
                "metadata": data.get("metadata") or data.get("metadata", {}),
            }
            return cls.model_validate(converted_data)
        elif isinstance(data, str):
            return cls.deserialize(json.loads(data))
        else:
            raise ValueError("Invalid span_context provided")


class ModelProvider(str, Enum):
    ANTHROPIC = "anthropic"
    BEDROCK = "bedrock"
    OPENAI = "openai"
    GEMINI = "gemini"


class MaskInputOptions(TypedDict):
    textarea: bool | None
    text: bool | None
    number: bool | None
    select: bool | None
    email: bool | None
    tel: bool | None


class SessionRecordingOptions(TypedDict):
    mask_input_options: MaskInputOptions | None



import json
import logging
from typing import Any, Dict

from .config import Config
from .utils import (
    JSONEncoder,
    dont_throw,
    model_as_dict,
    should_send_prompts,
    _extract_response_data,
    _aextract_response_data,
    set_span_attribute,
)
from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import (
    GEN_AI_RESPONSE_ID,
)
from opentelemetry.semconv_ai import SpanAttributes

logger = logging.getLogger(__name__)


def _is_base64_image(item: Dict[str, Any]) -> bool:
    if not isinstance(item, dict):
        return False

    if not isinstance(item.get("source"), dict):
        return False

    if item.get("type") != "image" or item["source"].get("type") != "base64":
        return False

    return True


async def _process_image_item(item, trace_id, span_id, message_index, content_index):
    if not Config.upload_base64_image:
        return item

    image_format = item.get("source").get("media_type").split("/")[1]
    image_name = f"message_{message_index}_content_{content_index}.{image_format}"
    base64_string = item.get("source").get("data")
    url = await Config.upload_base64_image(trace_id, span_id, image_name, base64_string)

    return {"type": "image_url", "image_url": {"url": url}}


async def _dump_content(message_index, content, span):
    if isinstance(content, str):
        return content
    elif isinstance(content, list):
        # If the content is a list of text blocks, concatenate them.
        # This is more commonly used in prompt caching.
        if all([model_as_dict(item).get("type") == "text" for item in content]):
            return "".join([model_as_dict(item).get("text") for item in content])

        content = [
            (
                await _process_image_item(
                    model_as_dict(item),
                    span.context.trace_id,
                    span.context.span_id,
                    message_index,
                    j,
                )
                if _is_base64_image(model_as_dict(item))
                else model_as_dict(item)
            )
            for j, item in enumerate(content)
        ]

        return json.dumps(content, cls=JSONEncoder)


@dont_throw
async def aset_input_attributes(span, kwargs):
    from .utils import set_span_attribute

    set_span_attribute(span, SpanAttributes.LLM_REQUEST_MODEL, kwargs.get("model"))
    set_span_attribute(
        span, SpanAttributes.LLM_REQUEST_MAX_TOKENS, kwargs.get("max_tokens_to_sample")
    )
    set_span_attribute(
        span, SpanAttributes.LLM_REQUEST_TEMPERATURE, kwargs.get("temperature")
    )
    set_span_attribute(span, SpanAttributes.LLM_REQUEST_TOP_P, kwargs.get("top_p"))
    set_span_attribute(
        span, SpanAttributes.LLM_FREQUENCY_PENALTY, kwargs.get("frequency_penalty")
    )
    set_span_attribute(
        span, SpanAttributes.LLM_PRESENCE_PENALTY, kwargs.get("presence_penalty")
    )
    set_span_attribute(span, SpanAttributes.LLM_IS_STREAMING, kwargs.get("stream"))
    set_span_attribute(
        span, "anthropic.request.service_tier", kwargs.get("service_tier")
    )

    if should_send_prompts():
        if kwargs.get("prompt") is not None:
            set_span_attribute(
                span, f"{SpanAttributes.LLM_PROMPTS}.0.user", kwargs.get("prompt")
            )

        elif kwargs.get("messages") is not None:
            has_system_message = False
            if kwargs.get("system"):
                has_system_message = True
                set_span_attribute(
                    span,
                    f"{SpanAttributes.LLM_PROMPTS}.0.content",
                    await _dump_content(
                        message_index=0, span=span, content=kwargs.get("system")
                    ),
                )
                set_span_attribute(
                    span,
                    f"{SpanAttributes.LLM_PROMPTS}.0.role",
                    "system",
                )
            for i, message in enumerate(kwargs.get("messages")):
                prompt_index = i + (1 if has_system_message else 0)
                content = message.get("content")
                tool_use_blocks = []
                other_blocks = []
                if isinstance(content, list):
                    for block in content:
                        if dict(block).get("type") == "tool_use":
                            tool_use_blocks.append(dict(block))
                        else:
                            other_blocks.append(block)
                    content = other_blocks
                set_span_attribute(
                    span,
                    f"{SpanAttributes.LLM_PROMPTS}.{prompt_index}.content",
                    await _dump_content(message_index=i, span=span, content=content),
                )
                set_span_attribute(
                    span,
                    f"{SpanAttributes.LLM_PROMPTS}.{prompt_index}.role",
                    message.get("role"),
                )
                if tool_use_blocks:
                    for tool_num, tool_use_block in enumerate(tool_use_blocks):
                        set_span_attribute(
                            span,
                            f"{SpanAttributes.LLM_PROMPTS}.{prompt_index}.tool_calls.{tool_num}.id",
                            tool_use_block.get("id"),
                        )
                        set_span_attribute(
                            span,
                            f"{SpanAttributes.LLM_PROMPTS}.{prompt_index}.tool_calls.{tool_num}.name",
                            tool_use_block.get("name"),
                        )
                        set_span_attribute(
                            span,
                            f"{SpanAttributes.LLM_PROMPTS}.{prompt_index}.tool_calls.{tool_num}.arguments",
                            json.dumps(tool_use_block.get("input")),
                        )

        if kwargs.get("tools") is not None:
            for i, tool in enumerate(kwargs.get("tools")):
                prefix = f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{i}"
                set_span_attribute(span, f"{prefix}.name", tool.get("name"))
                set_span_attribute(
                    span, f"{prefix}.description", tool.get("description")
                )
                input_schema = tool.get("input_schema")
                if input_schema is not None:
                    set_span_attribute(
                        span, f"{prefix}.input_schema", json.dumps(input_schema)
                    )


async def _aset_span_completions(span, response):
    if not should_send_prompts():
        return

    response = await _aextract_response_data(response)
    index = 0
    prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
    set_span_attribute(span, f"{prefix}.finish_reason", response.get("stop_reason"))
    if response.get("role"):
        set_span_attribute(span, f"{prefix}.role", response.get("role"))

    if response.get("completion"):
        set_span_attribute(span, f"{prefix}.content", response.get("completion"))
    elif response.get("content"):
        tool_call_index = 0
        text = ""
        for content in response.get("content"):
            content_block_type = content.type
            # usually, Antrhopic responds with just one text block,
            # but the API allows for multiple text blocks, so concatenate them
            if content_block_type == "text" and hasattr(content, "text"):
                text += content.text
            elif content_block_type == "thinking":
                content = dict(content)
                # override the role to thinking
                set_span_attribute(
                    span,
                    f"{prefix}.role",
                    "thinking",
                )
                set_span_attribute(
                    span,
                    f"{prefix}.content",
                    content.get("thinking"),
                )
                # increment the index for subsequent content blocks
                index += 1
                prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
                # set the role to the original role on the next completions
                set_span_attribute(
                    span,
                    f"{prefix}.role",
                    response.get("role"),
                )
            elif content_block_type == "tool_use":
                content = dict(content)
                set_span_attribute(
                    span,
                    f"{prefix}.tool_calls.{tool_call_index}.id",
                    content.get("id"),
                )
                set_span_attribute(
                    span,
                    f"{prefix}.tool_calls.{tool_call_index}.name",
                    content.get("name"),
                )
                tool_arguments = content.get("input")
                if tool_arguments is not None:
                    set_span_attribute(
                        span,
                        f"{prefix}.tool_calls.{tool_call_index}.arguments",
                        json.dumps(tool_arguments),
                    )
                tool_call_index += 1
        set_span_attribute(span, f"{prefix}.content", text)


def _set_span_completions(span, response):
    if not should_send_prompts():
        return
    from .utils import set_span_attribute

    index = 0
    prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
    set_span_attribute(span, f"{prefix}.finish_reason", response.get("stop_reason"))
    if response.get("role"):
        set_span_attribute(span, f"{prefix}.role", response.get("role"))

    if response.get("completion"):
        set_span_attribute(span, f"{prefix}.content", response.get("completion"))
    elif response.get("content"):
        tool_call_index = 0
        text = ""
        for content in response.get("content"):
            content_block_type = content.type
            # usually, Antrhopic responds with just one text block,
            # but the API allows for multiple text blocks, so concatenate them
            if content_block_type == "text" and hasattr(content, "text"):
                text += content.text or ""
            elif content_block_type == "thinking":
                content = dict(content)
                # override the role to thinking
                set_span_attribute(
                    span,
                    f"{prefix}.role",
                    "thinking",
                )
                set_span_attribute(
                    span,
                    f"{prefix}.content",
                    content.get("thinking"),
                )
                # increment the index for subsequent content blocks
                index += 1
                prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
                # set the role to the original role on the next completions
                set_span_attribute(
                    span,
                    f"{prefix}.role",
                    response.get("role"),
                )
            elif content_block_type == "tool_use":
                content = dict(content)
                set_span_attribute(
                    span,
                    f"{prefix}.tool_calls.{tool_call_index}.id",
                    content.get("id"),
                )
                set_span_attribute(
                    span,
                    f"{prefix}.tool_calls.{tool_call_index}.name",
                    content.get("name"),
                )
                tool_arguments = content.get("input")
                if tool_arguments is not None:
                    set_span_attribute(
                        span,
                        f"{prefix}.tool_calls.{tool_call_index}.arguments",
                        json.dumps(tool_arguments),
                    )
                tool_call_index += 1
        set_span_attribute(span, f"{prefix}.content", text)


@dont_throw
async def aset_response_attributes(span, response):
    response = await _aextract_response_data(response)
    set_span_attribute(span, SpanAttributes.LLM_RESPONSE_MODEL, response.get("model"))
    set_span_attribute(span, GEN_AI_RESPONSE_ID, response.get("id"))

    if usage := response.get("usage"):
        if hasattr(usage, "service_tier"):
            set_span_attribute(
                span,
                "anthropic.response.service_tier",
                usage.service_tier,
            )
        prompt_tokens = usage.input_tokens
        completion_tokens = usage.output_tokens
        set_span_attribute(span, SpanAttributes.LLM_USAGE_PROMPT_TOKENS, prompt_tokens)
        set_span_attribute(
            span, SpanAttributes.LLM_USAGE_COMPLETION_TOKENS, completion_tokens
        )
        set_span_attribute(
            span,
            SpanAttributes.LLM_USAGE_TOTAL_TOKENS,
            prompt_tokens + completion_tokens,
        )

    await _aset_span_completions(span, response)


@dont_throw
def set_response_attributes(span, response):
    response = _extract_response_data(response)
    set_span_attribute(span, SpanAttributes.LLM_RESPONSE_MODEL, response.get("model"))
    set_span_attribute(span, GEN_AI_RESPONSE_ID, response.get("id"))

    if usage := response.get("usage"):
        if hasattr(usage, "service_tier"):
            set_span_attribute(
                span,
                "anthropic.response.service_tier",
                usage.service_tier,
            )
        prompt_tokens = usage.input_tokens
        completion_tokens = usage.output_tokens
        set_span_attribute(span, SpanAttributes.LLM_USAGE_PROMPT_TOKENS, prompt_tokens)
        set_span_attribute(
            span, SpanAttributes.LLM_USAGE_COMPLETION_TOKENS, completion_tokens
        )
        set_span_attribute(
            span,
            SpanAttributes.LLM_USAGE_TOTAL_TOKENS,
            prompt_tokens + completion_tokens,
        )

    _set_span_completions(span, response)


@dont_throw
def set_streaming_response_attributes(span, complete_response_events):
    if not should_send_prompts():
        return

    if not span.is_recording() or not complete_response_events:
        return

    index = 0
    for event in complete_response_events:
        prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
        set_span_attribute(span, f"{prefix}.finish_reason", event.get("finish_reason"))
        role = "thinking" if event.get("type") == "thinking" else "assistant"
        # Thinking is added as a separate completion, so we need to increment the index
        if event.get("type") == "thinking":
            index += 1
        set_span_attribute(span, f"{prefix}.role", role)
        if event.get("type") == "tool_use":
            set_span_attribute(
                span,
                f"{prefix}.tool_calls.0.id",
                event.get("id"),
            )
            set_span_attribute(
                span,
                f"{prefix}.tool_calls.0.name",
                event.get("name"),
            )
            tool_arguments = event.get("input")
            if tool_arguments is not None:
                set_span_attribute(
                    span,
                    f"{prefix}.tool_calls.0.arguments",
                    # already stringified
                    tool_arguments,
                )
        else:
            set_span_attribute(span, f"{prefix}.content", event.get("text"))



# import base64
import base64
from copy import deepcopy
from typing import Any

from lmnr.sdk.utils import json_dumps
from pydantic import BaseModel


def screenshot_tool_output_formatter(output: Any) -> str:
    # output is of type BinaryAPIResponse, which implements
    # the iter_bytes method from httpx.Response

    return "<BINARY_BLOB_SCREENSHOT>"
    # The below implementation works, but it may consume the entire iterator,
    # making the response unusable after the formatter is called.
    # This is UNLESS somewhere in code output.read() (httpx.Response.read())
    # is called.
    # We cannot rely on that now, so we return a placeholder.
    # response_bytes = []
    # for chunk in output.iter_bytes():
    #     response_bytes.append(chunk)
    # response_base64 = base64.b64encode(response_bytes).decode("utf-8")
    # return f"data:image/png;base64,{response_base64}"


def process_tool_output_formatter(output: Any) -> str:
    if not isinstance(output, (dict, BaseModel)):
        return json_dumps(output)

    output = output.model_dump() if isinstance(output, BaseModel) else deepcopy(output)
    if "stderr_b64" in output:
        output["stderr"] = base64.b64decode(output["stderr_b64"]).decode("utf-8")
    if "stdout_b64" in output:
        output["stdout"] = base64.b64decode(output["stdout_b64"]).decode("utf-8")
    return json_dumps(output)



"""OpenTelemetry CUA instrumentation"""

import logging
from typing import Any, AsyncGenerator, Collection

from lmnr import Laminar
from lmnr.sdk.utils import json_dumps
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap

from opentelemetry.trace import Span
from opentelemetry.trace.status import Status, StatusCode
from wrapt import wrap_function_wrapper

logger = logging.getLogger(__name__)

_instruments = ("cua-agent >= 0.4.0",)


def _wrap_run(
    wrapped,
    instance,
    args,
    kwargs,
):
    parent_span = Laminar.start_span("ComputerAgent.run")
    instance._lmnr_parent_span = parent_span

    try:
        result: AsyncGenerator[dict[str, Any], None] = wrapped(*args, **kwargs)
        return _abuild_from_streaming_response(parent_span, result)
    except Exception as e:
        if parent_span.is_recording():
            parent_span.set_status(Status(StatusCode.ERROR))
            parent_span.record_exception(e)
            parent_span.end()
        raise


async def _abuild_from_streaming_response(
    parent_span: Span, response: AsyncGenerator[dict[str, Any], None]
) -> AsyncGenerator[dict[str, Any], None]:
    with Laminar.use_span(parent_span, end_on_exit=True):
        response_iter = aiter(response)
        while True:
            step = None
            step_span = Laminar.start_span("ComputerAgent.step")
            with Laminar.use_span(step_span):
                try:
                    step = await anext(response_iter)
                    step_span.set_attribute("lmnr.span.output", json_dumps(step))
                    try:
                        # When processing tool calls, each output item is processed separately,
                        # if the output is message, agent.step returns an empty array
                        # https://github.com/trycua/cua/blob/17d670962970a1d1774daaec029ebf92f1f9235e/libs/python/agent/agent/agent.py#L459
                        if len(step.get("output", [])) == 0:
                            continue
                    except Exception:
                        pass
                    if step_span.is_recording():
                        step_span.end()
                except StopAsyncIteration:
                    # don't end on purpose, there is no iteration step here.
                    break

            if step is not None:
                yield step


class CuaAgentInstrumentor(BaseInstrumentor):
    def __init__(self):
        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        wrap_package = "agent.agent"
        wrap_object = "ComputerAgent"
        wrap_method = "run"
        try:
            wrap_function_wrapper(
                wrap_package,
                f"{wrap_object}.{wrap_method}",
                _wrap_run,
            )
        except ModuleNotFoundError:
            pass  # that's ok, we don't want to fail if some methods do not exist

    def _uninstrument(self, **kwargs):
        wrap_package = "agent.agent"
        wrap_object = "ComputerAgent"
        wrap_method = "run"
        try:
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrap_method,
            )
        except ModuleNotFoundError:
            pass  # that's ok, we don't want to fail if some methods do not exist



from functools import wraps
import types
from typing import Any, AsyncGenerator, Callable, Generator, Literal, TypeVar

from opentelemetry import context as context_api
from opentelemetry.trace import Span, Status, StatusCode

from lmnr.opentelemetry_lib.tracing.context import (
    CONTEXT_METADATA_KEY,
    attach_context,
    detach_context,
    get_event_attributes_from_context,
)
from lmnr.opentelemetry_lib.tracing.span import LaminarSpan
from lmnr.opentelemetry_lib.tracing.utils import set_association_props_in_context
from lmnr.sdk.utils import get_input_from_func_args, is_method
from lmnr.opentelemetry_lib.tracing.tracer import get_tracer_with_context
from lmnr.opentelemetry_lib.tracing.attributes import (
    ASSOCIATION_PROPERTIES,
    METADATA,
    SPAN_TYPE,
)
from lmnr.opentelemetry_lib.tracing import TracerWrapper
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.utils import is_otel_attribute_value_type, json_dumps

logger = get_default_logger(__name__)

F = TypeVar("F", bound=Callable[..., Any])


def _setup_span(
    span_name: str,
    span_type: str,
    association_properties: dict[str, Any] | None,
    preserve_global_context: bool = False,
    metadata: dict[str, Any] | None = None,
):
    """Set up a span with the given name, type, and association properties."""
    with get_tracer_with_context() as (tracer, isolated_context):
        # Create span in isolated context
        span = tracer.start_span(
            span_name,
            context=isolated_context if not preserve_global_context else None,
            attributes={SPAN_TYPE: span_type},
        )

        ctx_metadata = context_api.get_value(CONTEXT_METADATA_KEY, isolated_context)
        merged_metadata = {
            **(ctx_metadata or {}),
            **(metadata or {}),
        }
        for key, value in merged_metadata.items():
            span.set_attribute(
                f"{ASSOCIATION_PROPERTIES}.{METADATA}.{key}",
                (value if is_otel_attribute_value_type(value) else json_dumps(value)),
            )

        if association_properties is not None:
            for key, value in association_properties.items():
                span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{key}", value)

        return span


def _process_input(
    span: Span,
    fn: Callable,
    args: tuple,
    kwargs: dict,
    ignore_input: bool,
    ignore_inputs: list[str] | None,
    input_formatter: Callable[..., str] | None,
):
    """Process and set input attributes on the span."""
    if ignore_input:
        return

    try:
        if input_formatter is not None:
            inp = input_formatter(*args, **kwargs)
        else:
            inp = get_input_from_func_args(
                fn,
                is_method=is_method(fn),
                func_args=args,
                func_kwargs=kwargs,
                ignore_inputs=ignore_inputs,
            )

        if not isinstance(span, LaminarSpan):
            span = LaminarSpan(span)
        span.set_input(inp)
    except Exception:
        msg = "Failed to process input, ignoring"
        if input_formatter is not None:
            # Only warn the user if they provided an input formatter
            # because it's their responsibility to make sure it works.
            logger.warning(msg, exc_info=True)
        else:
            logger.debug(msg, exc_info=True)
        pass


def _process_output(
    span: Span,
    result: Any,
    ignore_output: bool,
    output_formatter: Callable[..., str] | None,
):
    """Process and set output attributes on the span."""
    if ignore_output:
        return

    try:
        if output_formatter is not None:
            output = output_formatter(result)
        else:
            output = result

        if not isinstance(span, LaminarSpan):
            span = LaminarSpan(span)
        span.set_output(output)
    except Exception:
        msg = "Failed to process output, ignoring"
        if output_formatter is not None:
            # Only warn the user if they provided an output formatter
            # because it's their responsibility to make sure it works.
            logger.warning(msg, exc_info=True)
        else:
            logger.debug(msg, exc_info=True)
        pass


def _cleanup_span(span: Span, wrapper: TracerWrapper):
    """Clean up span and context."""
    span.end()
    wrapper.pop_span_context()


def observe_base(
    *,
    name: str | None = None,
    ignore_input: bool = False,
    ignore_inputs: list[str] | None = None,
    ignore_output: bool = False,
    span_type: Literal["DEFAULT", "LLM", "TOOL"] = "DEFAULT",
    metadata: dict[str, Any] | None = None,
    association_properties: dict[str, Any] | None = None,
    input_formatter: Callable[..., str] | None = None,
    output_formatter: Callable[..., str] | None = None,
    preserve_global_context: bool = False,
) -> Callable[[F], F]:
    def decorate(fn: F) -> F:
        @wraps(fn)
        def wrap(*args, **kwargs):
            if not TracerWrapper.verify_initialized():
                return fn(*args, **kwargs)

            span_name = name or fn.__name__
            wrapper = TracerWrapper()

            span = _setup_span(
                span_name,
                span_type,
                association_properties,
                preserve_global_context,
                metadata,
            )

            # Set association props in context before push_span_context
            # so child spans inherit them
            assoc_props_token = set_association_props_in_context(span)
            if assoc_props_token and isinstance(span, LaminarSpan):
                span._lmnr_assoc_props_token = assoc_props_token

            new_context = wrapper.push_span_context(span)
            # Some auto-instrumentations are not under our control, so they
            # don't have access to our isolated context. We attach the context
            # to the OTEL global context, so that spans know their parent
            # span and trace_id.
            ctx_token = context_api.attach(new_context)
            # update our isolated context too
            isolated_ctx_token = attach_context(new_context)

            _process_input(
                span, fn, args, kwargs, ignore_input, ignore_inputs, input_formatter
            )

            try:
                res = fn(*args, **kwargs)
            except Exception as e:
                _process_exception(span, e)
                _cleanup_span(span, wrapper)
                raise
            finally:
                # Always restore global context
                context_api.detach(ctx_token)
                detach_context(isolated_ctx_token)
            # span will be ended in the generator
            if isinstance(res, types.GeneratorType):
                return _handle_generator(span, wrapper, res)
            if isinstance(res, types.AsyncGeneratorType):
                # async def foo() -> AsyncGenerator[int, None]:
                # is not considered async in a classical sense in Python,
                # so we handle this inside the sync wrapper.
                # In particular, CO_COROUTINE is different from CO_ASYNC_GENERATOR.
                # Flags are listed from LSB here:
                # https://docs.python.org/3/library/inspect.html#inspect-module-co-flags
                # See also: https://groups.google.com/g/python-tulip/c/6rWweGXLutU?pli=1
                return _ahandle_generator(span, wrapper, res)

            _process_output(span, res, ignore_output, output_formatter)
            _cleanup_span(span, wrapper)
            return res

        return wrap

    return decorate


# Async Decorators
def async_observe_base(
    *,
    name: str | None = None,
    ignore_input: bool = False,
    ignore_inputs: list[str] | None = None,
    ignore_output: bool = False,
    span_type: Literal["DEFAULT", "LLM", "TOOL"] = "DEFAULT",
    metadata: dict[str, Any] | None = None,
    association_properties: dict[str, Any] | None = None,
    input_formatter: Callable[..., str] | None = None,
    output_formatter: Callable[..., str] | None = None,
    preserve_global_context: bool = False,
) -> Callable[[F], F]:
    def decorate(fn: F) -> F:
        @wraps(fn)
        async def wrap(*args, **kwargs):
            if not TracerWrapper.verify_initialized():
                return await fn(*args, **kwargs)

            span_name = name or fn.__name__
            wrapper = TracerWrapper()

            span = _setup_span(
                span_name,
                span_type,
                association_properties,
                preserve_global_context,
                metadata,
            )

            # Set association props in context before push_span_context
            # so child spans inherit them
            assoc_props_token = set_association_props_in_context(span)
            if assoc_props_token and isinstance(span, LaminarSpan):
                span._lmnr_assoc_props_token = assoc_props_token

            new_context = wrapper.push_span_context(span)
            # Some auto-instrumentations are not under our control, so they
            # don't have access to our isolated context. We attach the context
            # to the OTEL global context, so that spans know their parent
            # span and trace_id.
            ctx_token = context_api.attach(new_context)
            # update our isolated context too
            isolated_ctx_token = attach_context(new_context)

            _process_input(
                span, fn, args, kwargs, ignore_input, ignore_inputs, input_formatter
            )

            try:
                res = await fn(*args, **kwargs)
            except Exception as e:
                _process_exception(span, e)
                _cleanup_span(span, wrapper)
                raise e
            finally:
                # Always restore global context
                context_api.detach(ctx_token)
                detach_context(isolated_ctx_token)

            # span will be ended in the generator
            if isinstance(res, types.AsyncGeneratorType):
                # probably unreachable, read the comment in the similar
                # part of the sync wrapper.
                return await _ahandle_generator(span, wrapper, res)

            _process_output(span, res, ignore_output, output_formatter)
            _cleanup_span(span, wrapper)
            return res

        return wrap

    return decorate


def _handle_generator(
    span: Span,
    wrapper: TracerWrapper,
    res: Generator,
    ignore_output: bool = False,
    output_formatter: Callable[..., str] | None = None,
):
    results = []
    try:
        for part in res:
            results.append(part)
            yield part
    finally:
        _process_output(span, results, ignore_output, output_formatter)
        _cleanup_span(span, wrapper)


async def _ahandle_generator(
    span: Span,
    wrapper: TracerWrapper,
    res: AsyncGenerator,
    ignore_output: bool = False,
    output_formatter: Callable[..., str] | None = None,
):
    results = []
    try:
        async for part in res:
            results.append(part)
            yield part
    finally:
        _process_output(span, results, ignore_output, output_formatter)
        _cleanup_span(span, wrapper)


def _process_exception(span: Span, e: Exception):
    # Note that this `escaped` is sent as a StringValue("True"), not a boolean.
    span.record_exception(
        e, attributes=get_event_attributes_from_context(), escaped=True
    )
    span.set_status(Status(StatusCode.ERROR, str(e)))



import logging
import threading
import uuid

from opentelemetry.sdk.trace.export import (
    SpanProcessor,
    SpanExporter,
    BatchSpanProcessor,
    SimpleSpanProcessor,
)
from opentelemetry.sdk.trace import Span
from opentelemetry.context import Context, get_value

from lmnr.opentelemetry_lib.tracing.attributes import (
    ASSOCIATION_PROPERTIES,
    PARENT_SPAN_IDS_PATH,
    PARENT_SPAN_PATH,
    SESSION_ID,
    SPAN_IDS_PATH,
    SPAN_INSTRUMENTATION_SOURCE,
    SPAN_LANGUAGE_VERSION,
    SPAN_PATH,
    SPAN_SDK_VERSION,
    TRACE_TYPE,
    USER_ID,
)
from lmnr.opentelemetry_lib.tracing.context import (
    CONTEXT_METADATA_KEY,
    CONTEXT_SESSION_ID_KEY,
    CONTEXT_TRACE_TYPE_KEY,
    CONTEXT_USER_ID_KEY,
)
from lmnr.opentelemetry_lib.tracing.exporter import LaminarSpanExporter
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.utils import from_env, is_otel_attribute_value_type, json_dumps
from lmnr.version import PYTHON_VERSION, __version__


class LaminarSpanProcessor(SpanProcessor):
    instance: BatchSpanProcessor | SimpleSpanProcessor
    logger: logging.Logger
    __span_id_to_path: dict[int, list[str]] = {}
    __span_id_lists: dict[int, list[str]] = {}
    max_export_batch_size: int
    _instance_lock: threading.RLock
    _paths_lock: threading.RLock

    def __init__(
        self,
        base_url: str | None = None,
        port: int | None = None,
        api_key: str | None = None,
        timeout_seconds: int = 30,
        force_http: bool = False,
        max_export_batch_size: int = 64,
        disable_batch: bool = False,
        exporter: SpanExporter | None = None,
    ):
        self._instance_lock = threading.RLock()
        self._paths_lock = threading.RLock()
        self.logger = get_default_logger(__name__)
        self.max_export_batch_size = max_export_batch_size
        self.exporter = exporter or LaminarSpanExporter(
            base_url=base_url,
            port=port,
            api_key=api_key,
            timeout_seconds=timeout_seconds,
            force_http=force_http,
        )
        self.instance = (
            SimpleSpanProcessor(self.exporter)
            if disable_batch
            else BatchSpanProcessor(
                self.exporter, max_export_batch_size=max_export_batch_size
            )
        )

    def on_start(self, span: Span, parent_context: Context | None = None):
        is_disabled = (
            from_env("LMNR_DISABLE_TRACING") or "false"
        ).lower().strip() == "true"

        if is_disabled:
            span.set_attribute("lmnr.internal.disabled", True)

        with self._paths_lock:
            parent_span_path = list(span.attributes.get(PARENT_SPAN_PATH, tuple())) or (
                self.__span_id_to_path.get(span.parent.span_id) if span.parent else None
            )
            parent_span_ids_path = list(
                span.attributes.get(PARENT_SPAN_IDS_PATH, tuple())
            ) or (
                self.__span_id_lists.get(span.parent.span_id, []) if span.parent else []
            )
            span_name_in_path = span.name if not is_disabled else "_"
            span_path = (
                parent_span_path + [span_name_in_path]
                if parent_span_path
                else [span_name_in_path]
            )
            span_ids_path = parent_span_ids_path + [
                str(uuid.UUID(int=span.get_span_context().span_id))
            ]
            span.set_attribute(SPAN_PATH, span_path)
            span.set_attribute(SPAN_IDS_PATH, span_ids_path)
            self.__span_id_to_path[span.get_span_context().span_id] = span_path
            self.__span_id_lists[span.get_span_context().span_id] = span_ids_path

        if is_disabled:
            return

        span.set_attribute(SPAN_INSTRUMENTATION_SOURCE, "python")
        span.set_attribute(SPAN_SDK_VERSION, __version__)
        span.set_attribute(SPAN_LANGUAGE_VERSION, f"python@{PYTHON_VERSION}")

        if parent_context:
            trace_type = get_value(CONTEXT_TRACE_TYPE_KEY, parent_context)
            if trace_type:
                span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{TRACE_TYPE}", trace_type)
            user_id = get_value(CONTEXT_USER_ID_KEY, parent_context)
            if user_id:
                span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{USER_ID}", user_id)
            session_id = get_value(CONTEXT_SESSION_ID_KEY, parent_context)
            if session_id:
                span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{SESSION_ID}", session_id)
            ctx_metadata = get_value(CONTEXT_METADATA_KEY, parent_context)
            if ctx_metadata and isinstance(ctx_metadata, dict):
                span_metadata = {}
                if hasattr(span, "attributes") and hasattr(span.attributes, "items"):
                    for key, value in span.attributes.items():
                        if key.startswith(f"{ASSOCIATION_PROPERTIES}.metadata."):
                            span_metadata[
                                key.replace(f"{ASSOCIATION_PROPERTIES}.metadata.", "")
                            ] = value

                for key, value in {**ctx_metadata, **span_metadata}.items():
                    span.set_attribute(
                        f"{ASSOCIATION_PROPERTIES}.metadata.{key}",
                        (
                            value
                            if is_otel_attribute_value_type(value)
                            else json_dumps(value)
                        ),
                    )

        if span.name == "LangGraph.workflow":
            graph_context = get_value("lmnr.langgraph.graph") or {}
            for key, value in graph_context.items():
                span.set_attribute(f"lmnr.association.properties.{key}", value)

        with self._instance_lock:
            self.instance.on_start(span, parent_context)

    def on_end(self, span: Span):
        if (from_env("LMNR_DISABLE_TRACING") or "false").lower().strip() == "true" or (
            span.attributes and span.attributes.get("lmnr.internal.disabled")
        ):
            return
        with self._instance_lock:
            self.instance.on_end(span)

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        with self._instance_lock:
            return self.instance.force_flush(timeout_millis)

    def force_reinit(self):
        if not isinstance(self.exporter, LaminarSpanExporter):
            self.logger.warning(
                "LaminarSpanProcessor is not using LaminarSpanExporter, cannot force reinit"
            )
            return

        # Reinitialize exporter (thread-safe, handles its own locking)
        self.exporter._init_instance()

        with self._instance_lock:
            old_instance = self.instance
            disable_batch = isinstance(old_instance, SimpleSpanProcessor)

            try:
                old_instance.shutdown()
            except Exception as e:
                self.logger.debug(f"Error shutting down old processor instance: {e}")

            self.instance = (
                SimpleSpanProcessor(self.exporter)
                if disable_batch
                else BatchSpanProcessor(
                    self.exporter, max_export_batch_size=self.max_export_batch_size
                )
            )

    def shutdown(self):
        with self._instance_lock:
            self.instance.shutdown()

    def clear(self):
        with self._paths_lock:
            self.__span_id_to_path = {}
            self.__span_id_lists = {}

    def set_parent_path_info(
        self,
        parent_span_id: int,
        span_path: list[str],
        span_ids_path: list[str],
    ):
        with self._paths_lock:
            self.__span_id_to_path[parent_span_id] = span_path
            self.__span_id_lists[parent_span_id] = span_ids_path



from typing import Collection

from lmnr import Laminar
from lmnr.opentelemetry_lib.tracing.context import get_current_context
from lmnr.sdk.log import get_default_logger

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.trace import NonRecordingSpan, get_current_span
from wrapt import wrap_function_wrapper


_instruments = ("bubus >= 1.3.0",)
event_id_to_span_context = {}
logger = get_default_logger(__name__)


def wrap_dispatch(wrapped, instance, args, kwargs):
    event = args[0] if args and len(args) > 0 else kwargs.get("event", None)
    if event and hasattr(event, "event_id"):
        event_id = event.event_id
        if event_id:
            span = get_current_span(get_current_context())
            event_id_to_span_context[event_id] = span.get_span_context()
    return wrapped(*args, **kwargs)


async def wrap_process_event(wrapped, instance, args, kwargs):
    event = args[0] if args and len(args) > 0 else kwargs.get("event", None)
    span_context = None
    if event and hasattr(event, "event_id"):
        event_id = event.event_id
        if event_id:
            span_context = event_id_to_span_context.get(event_id)
    if not span_context:
        return await wrapped(*args, **kwargs)
    if not Laminar.is_initialized():
        return await wrapped(*args, **kwargs)
    with Laminar.use_span(NonRecordingSpan(span_context)):
        return await wrapped(*args, **kwargs)


class BubusInstrumentor(BaseInstrumentor):
    def __init__(self):
        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        try:
            wrap_function_wrapper("bubus.service", "EventBus.dispatch", wrap_dispatch)
        except (ModuleNotFoundError, ImportError):
            pass
        try:
            wrap_function_wrapper(
                "bubus.service", "EventBus.process_event", wrap_process_event
            )
        except (ModuleNotFoundError, ImportError):
            pass

    def _uninstrument(self, **kwargs):
        try:
            unwrap("bubus.service", "EventBus.dispatch")
        except (ModuleNotFoundError, ImportError):
            pass
        try:
            unwrap("bubus.service", "EventBus.process_event")
        except (ModuleNotFoundError, ImportError):
            pass
        event_id_to_span_context.clear()



import datetime
import dataclasses
import dotenv
import enum
import inspect
import os
import orjson
import pydantic
import queue
import typing
import uuid

from lmnr.sdk.log import get_default_logger

logger = get_default_logger(__name__)


def is_method(func: typing.Callable) -> bool:
    # inspect.ismethod is True for bound methods only, but in the decorator,
    # the method is not bound yet, so we need to check if the first parameter
    # is either 'self' or 'cls'. This only relies on naming conventions

    # `signature._parameters` is an OrderedDict,
    # so the order of insertion is preserved
    params = list(inspect.signature(func).parameters)
    return len(params) > 0 and params[0] in ["self", "cls"]


def is_async(func: typing.Callable) -> bool:
    # `__wrapped__` is set automatically by `functools.wraps` and
    # `functools.update_wrapper`
    # so we can use it to get the original function
    while hasattr(func, "__wrapped__"):
        func = func.__wrapped__

    if not inspect.isfunction(func):
        return False

    # Check if the function is asynchronous
    if inspect.iscoroutinefunction(func):
        return True

    # Fallback: check if the function's code object contains 'async'.
    # This is for cases when a decorator (not ours) did not properly use
    # `functools.wraps` or `functools.update_wrapper`
    return (func.__code__.co_flags & inspect.CO_COROUTINE) != 0


def is_async_iterator(o: typing.Any) -> bool:
    return hasattr(o, "__aiter__") and hasattr(o, "__anext__")


def is_iterator(o: typing.Any) -> bool:
    return hasattr(o, "__iter__") and hasattr(o, "__next__")


def serialize(obj: typing.Any) -> str | dict[str, typing.Any]:
    def serialize_inner(o: typing.Any):
        if isinstance(o, (datetime.datetime, datetime.date)):
            return o.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
        elif o is None:
            return None
        elif isinstance(o, (int, float, str, bool)):
            return o
        elif isinstance(o, uuid.UUID):
            return str(o)  # same as in final return, but explicit
        elif isinstance(o, enum.Enum):
            return o.value
        elif dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        elif isinstance(o, bytes):
            return o.decode("utf-8")
        elif isinstance(o, pydantic.BaseModel):
            return o.model_dump()
        elif isinstance(o, (tuple, set, frozenset)):
            return [serialize_inner(item) for item in o]
        elif isinstance(o, list):
            return [serialize_inner(item) for item in o]
        elif isinstance(o, dict):
            return {serialize_inner(k): serialize_inner(v) for k, v in o.items()}
        elif isinstance(o, queue.Queue):
            return type(o).__name__

        return str(o)

    return serialize_inner(obj)


def get_input_from_func_args(
    func: typing.Callable,
    is_method: bool = False,
    func_args: list[typing.Any] = [],
    func_kwargs: dict[str, typing.Any] = {},
    ignore_inputs: list[str] | None = None,
) -> dict[str, typing.Any]:
    # Remove implicitly passed "self" or "cls" argument for
    # instance or class methods
    res = {
        k: v
        for k, v in func_kwargs.items()
        if not (ignore_inputs and k in ignore_inputs)
    }
    for i, k in enumerate(inspect.signature(func).parameters.keys()):
        if is_method and k in ["self", "cls"]:
            continue
        if ignore_inputs and k in ignore_inputs:
            continue
        # If param has default value, then it's not present in func args
        if i < len(func_args):
            res[k] = func_args[i]
    return res


def from_env(key: str) -> str | None:
    if val := os.getenv(key):
        return val
    dotenv_path = dotenv.find_dotenv(usecwd=True)
    # use DotEnv directly so we can set verbose to False
    return dotenv.main.DotEnv(dotenv_path, verbose=False, encoding="utf-8").get(key)


def is_otel_attribute_value_type(value: typing.Any) -> bool:
    def is_primitive_type(value: typing.Any) -> bool:
        return isinstance(value, (int, float, str, bool))

    if is_primitive_type(value):
        return True
    elif isinstance(value, typing.Sequence):
        if len(value) > 0:
            return is_primitive_type(value[0]) and all(
                isinstance(v, type(value[0])) for v in value
            )
        return True
    return False


def get_otel_env_var(var_name: str) -> str | None:
    """Get OTEL environment variable with priority order.

    Checks in order:
    1. OTEL_EXPORTER_OTLP_TRACES_{var_name}
    2. OTEL_EXPORTER_OTLP_{var_name}
    3. OTEL_{var_name}

    Args:
        var_name: The variable name (e.g., 'ENDPOINT', 'HEADERS', 'TIMEOUT')

    Returns:
        str | None: The environment variable value or None if not found
    """
    candidates = [
        f"OTEL_EXPORTER_OTLP_TRACES_{var_name}",
        f"OTEL_EXPORTER_OTLP_{var_name}",
        f"OTEL_{var_name}",
    ]

    for candidate in candidates:
        if value := from_env(candidate):
            return value
    return None


def parse_otel_headers(headers_str: str | None) -> dict[str, str]:
    """Parse OTEL headers string into dictionary.

    Format: key1=value1,key2=value2
    Values are URL-decoded.

    Args:
        headers_str: Headers string in OTEL format

    Returns:
        dict[str, str]: Parsed headers dictionary
    """
    if not headers_str:
        return {}

    headers = {}
    for pair in headers_str.split(","):
        if "=" in pair:
            key, value = pair.split("=", 1)
            import urllib.parse

            headers[key.strip()] = urllib.parse.unquote(value.strip())
    return headers


def format_id(id_value: str | int | uuid.UUID) -> str:
    """Format trace/span/evaluation ID to a UUID string, or return valid UUID strings as-is.

    Args:
        id_value: The ID in various formats (UUID, int, or valid UUID string)

    Returns:
        str: UUID string representation

    Raises:
        ValueError: If id_value cannot be converted to a valid UUID
    """
    if isinstance(id_value, uuid.UUID):
        return str(id_value)
    elif isinstance(id_value, int):
        return str(uuid.UUID(int=id_value))
    elif isinstance(id_value, str):
        uuid.UUID(id_value)
        return id_value
    else:
        raise ValueError(f"Invalid ID type: {type(id_value)}")


DEFAULT_PLACEHOLDER = {}


def default_json(o):
    if isinstance(o, pydantic.BaseModel):
        return o.model_dump()

    # Handle various sequence types, but not strings or bytes
    if isinstance(o, (list, tuple, set, frozenset)):
        return list(o)

    try:
        return str(o)
    except Exception:
        logger.debug("Failed to serialize data to JSON, inner type: %s", type(o))
        pass
    return DEFAULT_PLACEHOLDER


def json_dumps(data: dict) -> str:
    try:
        return orjson.dumps(
            data,
            default=default_json,
            option=orjson.OPT_SERIALIZE_DATACLASS
            | orjson.OPT_SERIALIZE_UUID
            | orjson.OPT_UTC_Z
            | orjson.OPT_NON_STR_KEYS,
        ).decode("utf-8")
    except Exception:
        # Log the exception and return a placeholder if serialization completely fails
        logger.info("Failed to serialize data to JSON, type: %s", type(data))
        return "{}"  # Return an empty JSON object as a fallback



import base64
from collections import defaultdict
import logging
import traceback
from typing_extensions import TypedDict

from .config import (
    Config,
)
from google.genai import types
from google.genai._common import BaseModel
import pydantic
from opentelemetry.trace import Span
from typing import Any, Literal


class ToolCall(pydantic.BaseModel):
    name: str | None = pydantic.Field(default=None)
    id: str | None = pydantic.Field(default=None)
    arguments: dict[str, Any] = pydantic.Field(default={})


class ImageUrlInner(pydantic.BaseModel):
    url: str = pydantic.Field(default="")


class ImageUrl(pydantic.BaseModel):
    type: Literal["image_url"] = pydantic.Field(default="image_url")
    image_url: ImageUrlInner = pydantic.Field(default=ImageUrlInner())


class ProcessedContentPart(pydantic.BaseModel):
    content: str | None = pydantic.Field(default=None)
    function_call: ToolCall | None = pydantic.Field(default=None)
    image_url: ImageUrl | None = pydantic.Field(default=None)


class ProcessChunkResult(TypedDict):
    role: str
    model_version: str | None


def merge_text_parts(
    parts: list[types.PartDict | types.File | types.Part | str],
) -> list[types.Part]:
    if not parts:
        return []

    merged_parts: list[types.Part] = []
    accumulated_text = ""

    for part in parts:
        # Handle string input - treat as text
        if isinstance(part, str):
            accumulated_text += part
        # Handle File objects - they are not text, so don't merge
        elif isinstance(part, types.File):
            # Flush any accumulated text first
            if accumulated_text:
                merged_parts.append(types.Part(text=accumulated_text))
                accumulated_text = ""
            # Add the File as-is (wrapped in a Part if needed)
            # Note: File objects should be passed through as-is in the original part
            merged_parts.append(part)
        # Handle Part and PartDict (dicts)
        else:
            part_dict = to_dict(part)

            # Check if this is a text part
            if part_dict.get("text") is not None:
                accumulated_text += part_dict.get("text")
            else:
                # Non-text part (inline_data, function_call, etc.)
                # Flush any accumulated text first
                if accumulated_text:
                    merged_parts.append(types.Part(text=accumulated_text))
                    accumulated_text = ""

                # Add the non-text part as-is
                if isinstance(part, types.Part):
                    merged_parts.append(part)
                elif isinstance(part, dict):
                    # Convert dict to Part object
                    merged_parts.append(types.Part(**part_dict))

    # Don't forget to add any remaining accumulated text
    if accumulated_text:
        merged_parts.append(types.Part(text=accumulated_text))

    return merged_parts


def set_span_attribute(span: Span, name: str, value: Any):
    if value is not None and value != "":
        span.set_attribute(name, value)
    return


def dont_throw(func):
    """
    A decorator that wraps the passed in function and logs exceptions instead of throwing them.

    @param func: The function to wrap
    @return: The wrapper function
    """
    # Obtain a logger specific to the function's module
    logger = logging.getLogger(func.__module__)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.debug(
                "Laminar failed to trace in %s, error: %s",
                func.__name__,
                traceback.format_exc(),
            )
            if Config.exception_logger:
                Config.exception_logger(e)

    return wrapper


def to_dict(obj: BaseModel | pydantic.BaseModel | dict) -> dict[str, Any]:
    try:
        if isinstance(obj, BaseModel):
            return obj.model_dump()
        elif isinstance(obj, pydantic.BaseModel):
            return obj.model_dump()
        elif isinstance(obj, dict):
            return obj
        else:
            return dict(obj)
    except Exception as e:
        logging.error(f"Error converting to dict: {obj}, error: {e}")
        return dict(obj)


def get_content(
    content: (
        ProcessedContentPart | dict | list[ProcessedContentPart | dict] | str | None
    ),
) -> dict | list[dict] | None:
    if isinstance(content, dict):
        return content.get("content") or content.get("image_url")
    if isinstance(content, ProcessedContentPart):
        if content.content and isinstance(content.content, str):
            return {
                "type": "text",
                "text": content.content,
            }
        elif content.image_url:
            return content.image_url.model_dump()
        else:
            return None
    elif isinstance(content, list):
        contents_list = [get_content(item) for item in content]
        return [item for item in contents_list if item is not None]
    elif isinstance(content, str):
        return {
            "type": "text",
            "text": content,
        }
    else:
        return None


def process_content_union(
    content: types.ContentUnion | types.ContentUnionDict,
) -> ProcessedContentPart | dict | list[ProcessedContentPart | dict] | None:
    if isinstance(content, types.Content):
        parts = to_dict(content).get("parts", [])
        return [_process_part(part) for part in parts]
    elif isinstance(content, list):
        return [_process_part_union(item) for item in content]
    elif isinstance(content, (types.Part, types.File, str)):
        return _process_part_union(content)
    elif isinstance(content, dict):
        if "parts" in content:
            return [_process_part_union(item) for item in content.get("parts", [])]
        else:
            # Assume it's PartDict
            return _process_part_union(content)
    else:
        return None


def _process_part_union(
    content: types.PartDict | types.File | types.Part | str,
) -> ProcessedContentPart | dict | None:
    if isinstance(content, str):
        return ProcessedContentPart(content=content)
    elif isinstance(content, types.File):
        content_dict = to_dict(content)
        name = (
            content_dict.get("name")
            or content_dict.get("display_name")
            or content_dict.get("uri")
        )
        return ProcessedContentPart(content=f"files/{name}")
    elif isinstance(content, (types.Part, dict)):
        return _process_part(content)
    else:
        return None


def _process_part(
    content: types.Part,
) -> ProcessedContentPart | dict | None:
    part_dict = to_dict(content)
    if part_dict.get("inline_data"):
        blob = to_dict(part_dict.get("inline_data"))
        if blob.get("mime_type", "").startswith("image/"):
            return _process_image_item(blob)
        else:
            # currently, only images are supported
            return ProcessedContentPart(
                content=blob.get("mime_type") or "unknown_media"
            )
    elif function_call := part_dict.get("function_call"):
        function_call_dict = to_dict(function_call)
        return ProcessedContentPart(
            function_call=ToolCall(
                name=function_call_dict.get("name"),
                id=function_call_dict.get("id"),
                arguments=function_call_dict.get("args", {}),
            )
        )
    elif part_dict.get("text") is not None:
        return ProcessedContentPart(content=part_dict.get("text"))
    else:
        return None


def role_from_content_union(
    content: types.ContentUnion | types.ContentUnionDict,
) -> str | None:
    role = None
    if isinstance(content, types.Content):
        role = to_dict(content).get("role")
    elif isinstance(content, list) and len(content) > 0:
        role = role_from_content_union(content[0])
    elif isinstance(content, dict):
        role = content.get("role")
    else:
        return None
    return role
    # return "assistant" if role == "model" else role


def with_tracer_wrapper(func):
    """Helper for providing tracer for wrapper functions."""

    def _with_tracer(tracer, to_wrap):
        def wrapper(wrapped, instance, args, kwargs):
            return func(tracer, to_wrap, wrapped, instance, args, kwargs)

        return wrapper

    return _with_tracer


def _process_image_item(blob: dict[str, Any]) -> ProcessedContentPart | dict | None:
    # Convert to openai format, so backends can handle it
    data = blob.get("data")
    encoded_data = (
        base64.b64encode(data).decode("utf-8") if isinstance(data, bytes) else data
    )
    mime_type = blob.get("mime_type", "image/unknown")
    image_type = mime_type.split("/")[1] if "/" in mime_type else "unknown"

    return (
        ProcessedContentPart(
            image_url=ImageUrl(
                image_url=ImageUrlInner(
                    url=f"data:image/{image_type};base64,{encoded_data}",
                )
            )
        )
        if Config.convert_image_to_openai_format
        else blob
    )


@dont_throw
def process_stream_chunk(
    chunk: types.GenerateContentResponse,
    existing_role: str,
    existing_model_version: str | None,
    # ============================== #
    # mutable states, passed by reference
    aggregated_usage_metadata: defaultdict[str, int],
    final_parts: list[types.Part | None],
    # ============================== #
) -> ProcessChunkResult:
    role = existing_role
    model_version = existing_model_version

    if chunk.model_version:
        model_version = chunk.model_version

    # Currently gemini throws an error if you pass more than one candidate
    # with streaming
    if chunk.candidates and len(chunk.candidates) > 0 and chunk.candidates[0].content:
        final_parts += chunk.candidates[0].content.parts or []
        role = chunk.candidates[0].content.role or role
    if chunk.usage_metadata:
        usage_dict = to_dict(chunk.usage_metadata)
        # prompt token count is sent in every chunk
        # (and is less by 1 in the last chunk, so we set it once);
        # total token count in every chunk is greater by prompt token count than it should be,
        # thus this awkward logic here
        if aggregated_usage_metadata.get("prompt_token_count") is None:
            # or 0, not .get(key, 0), because sometimes the value is explicitly None
            aggregated_usage_metadata["prompt_token_count"] = (
                usage_dict.get("prompt_token_count") or 0
            )
            aggregated_usage_metadata["total_token_count"] = (
                usage_dict.get("total_token_count") or 0
            )
        aggregated_usage_metadata["candidates_token_count"] += (
            usage_dict.get("candidates_token_count") or 0
        )
        aggregated_usage_metadata["total_token_count"] += (
            usage_dict.get("candidates_token_count") or 0
        )
    return ProcessChunkResult(
        role=role,
        model_version=model_version,
    )
